from __future__ import print_function, unicode_literals

import pytest
from django import forms

from django_auxilium.forms.range import RangeSelectorField


class TestRangeSelectorField(object):
    """
    Test the rangeconfig selector validation
    """

    def setup_method(self, method):
        self.field = RangeSelectorField(required=False)

    def test_empty(self):
        data = ['', None]
        for d in data:
            assert self.field.clean(d) is None

    def test_invalid_input(self):
        data = [':', 'A', '1', 'A:1', 'B:A', '2:1', 'BB:AA']
        for d in data:
            with pytest.raises(forms.ValidationError):
                self.field.clean(d)

    def test_columns(self):
        data = [
            ('A:A', (1, None, 1, None)),
            ('A:B', (1, None, 2, None)),
            ('AA:BB', (27, None, 54, None))
        ]
        for d in data:
            assert self.field.clean(d[0]) == d[1]

    def test_rows(self):
        data = [
            ('1:1', (None, 1, None, 1)),
            ('1:2', (None, 1, None, 2)),
            ('11:22', (None, 11, None, 22))
        ]
        for d in data:
            assert self.field.clean(d[0]) == d[1]

    def test_both(self):
        data = [
            ('A1:A1', (1, 1, 1, 1)),
            ('A1:B2', (1, 1, 2, 2)),
            ('AA11:BB22', (27, 11, 54, 22))
        ]
        for d in data:
            assert self.field.clean(d[0]) == d[1]

    def test_max_rows(self):
        field = RangeSelectorField(max_rows=2)
        valid_data = [
            ('1:1', (None, 1, None, 1)),
            ('1:2', (None, 1, None, 2)),
            ('2:3', (None, 2, None, 3)),
        ]
        invalid_data = ['1:3', '11:22']

        for d in valid_data:
            assert field.clean(d[0]) == d[1]
        for i in invalid_data:
            with pytest.raises(forms.ValidationError):
                field.clean(i)

    def test_max_columns(self):
        field = RangeSelectorField(max_columns=2)
        valid_data = [
            ('A:A', (1, None, 1, None)),
            ('A:B', (1, None, 2, None)),
            ('B:C', (2, None, 3, None)),
        ]
        invalid_data = ['A:C', 'AA:BB']

        for d in valid_data:
            assert field.clean(d[0]) == d[1]
        for i in invalid_data:
            with pytest.raises(forms.ValidationError):
                field.clean(i)

    def test_max_both(self):
        field = RangeSelectorField(max_rows=2, max_columns=2)
        valid_data = [
            ('A1:A1', (1, 1, 1, 1)),
            ('A1:B2', (1, 1, 2, 2)),
            ('B2:C3', (2, 2, 3, 3)),
        ]
        invalid_data = ['A1:C3', 'AA11:BB22']
        for d in valid_data:
            assert field.clean(d[0]) == d[1]
        for i in invalid_data:
            with pytest.raises(forms.ValidationError):
                field.clean(i)

    def test_max_either(self):
        field = RangeSelectorField(max_either=2)
        valid_data = [
            ('A:A', (1, None, 1, None)),
            ('1:1', (None, 1, None, 1)),
            ('A1:A50', (1, 1, 1, 50)),
            ('A1:AA1', (1, 1, 27, 1)),
        ]
        invalid_data = ['A1:C3', 'AA11:BB22']
        for d in valid_data:
            assert field.clean(d[0]) == d[1]
        for i in invalid_data:
            with pytest.raises(forms.ValidationError):
                field.clean(i)

    def test_required_columns(self):
        field = RangeSelectorField(required_columns=True)

        with pytest.raises(forms.ValidationError):
            field.clean('1:5')

    def test_required_rows(self):
        field = RangeSelectorField(required_rows=True)

        with pytest.raises(forms.ValidationError):
            field.clean('A:C')



"""
Collection of simple abstract base models with a few additional attributes
which don't require much logic.
"""

from __future__ import print_function, unicode_literals

from django.conf import settings
from django.db import models
from django.utils.encoding import python_2_unicode_compatible


User = settings.AUTH_USER_MODEL


class CreatedModel(models.Model):
    """
    This model adds date created attribute

    Attributes
    ----------
    created : DateTimeField
        The datetime when the model instance is created
    """
    created = models.DateTimeField(auto_now_add=True)

    class Meta(object):
        abstract = True


class ModifiedModel(models.Model):
    """
    This model adds the last date modified attribute

    Attributes
    ----------
    modified : DateTimeField
        The datetime when the model was last modified and saved
    """
    modified = models.DateTimeField(auto_now=True)

    class Meta(object):
        abstract = True


class BaseModel(CreatedModel, ModifiedModel):
    """
    This model adds date created and last date modified attributes
    """

    class Meta(object):
        abstract = True


class UserModel(models.Model):
    """
    This model adds a foreign key to a user attribute

    Attributes
    ----------
    user : User
        The user associated with the model
    """
    user = models.ForeignKey(User)

    class Meta(object):
        abstract = True


class NoteModel(BaseModel):
    """
    This model adds a notes field

    Attributes
    ----------
    notes : TextField
        Text field for storing notes
    """
    notes = models.TextField(blank=True)

    class Meta:
        abstract = True


@python_2_unicode_compatible
class TitleDescriptionModel(BaseModel):
    """
    This model adds a title and description fields

    Attributes
    ----------
    title : CharField
        Maximum length is 256 character
    description : TextField
    """
    title = models.CharField(max_length=256)
    description = models.TextField(blank=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.title



from __future__ import absolute_import, print_function, unicode_literals

import six

from django_auxilium.models.base import TitleDescriptionModel


class TestTitleDescriptionModel(object):
    def test_str(self):
        m = TitleDescriptionModel(title='foo')

        assert six.text_type(m) == 'foo'



from __future__ import absolute_import, print_function, unicode_literals

from django import forms

from django_auxilium.views.wizard import WizardView


class _TestForm(forms.Form):
    pass


class _TestWizardView(WizardView):
    form_list = [_TestForm]


def test_wizard_view():
    assert _TestWizardView.get_initkwargs()['form_list'] == {
        '0': _TestForm,
    }



from __future__ import print_function, unicode_literals
import re

import six
from django import forms
from django.utils.translation import ugettext_lazy as _


retype = type(re.compile(''))


class MultipleValuesCharField(forms.CharField):
    """
    Form field to allow users to enter multiple values.

    The best approach to enter multiple values is to provide a user with multiple
    input fields however sometimes that is not feasible. In that case this field
    provides a way for a user to enter multiple values in a single input field. The
    values can be delimited by either a constant character(s) or even by a regular
    expression.

    Examples
    --------

    ::

        MultipleValuesCharField(delimiter=',')
        MultipleValuesCharField(delimiter=',', min_values=1, max_values=10)
        MultipleValuesCharField(delimiter=re.compile(r'\W+'), separator='\\n')
        MultipleValuesCharField(mapping=int)
        MultipleValuesCharField(mapping=lambda i: i == 5)
        MultipleValuesCharField(mapping={'foo': 'bar'})

    Parameters
    ----------
    delimiter : str, Regex, optional
        The delimiter according to which the values are split. It can also be given as a
        compiled regular expression (e.g. ``re.compile('\W+')``).
        Default is ``','`` (comma).
    separator : str
        Iff the delimiter is a regular expression, then this value will be used to
        separate values within the widget when rendering values back in the UI.
    mapping : dict, callable, optional
        By default all split values are casted to a Python string. The mapping allows to
        change that so that individual values will be mapped to different data-types.
        Mapping can be defined as a callable which will should accept one parameter, the
        value, and return the input value properly casted. Mapping can also be defined
        as a dictionary. In that case, if the individual value exists as a key of the
        dictionary, the value associated with the key will be used as a final casted
        value. Please note that if the key is not found in the dictionary, then the input
        value will be used.
        Default is ``None``.
    max_values : int, optional
        The maximum allowed number of values. Default is ``None``.
    min_values : int, optional
        The minimum required number of provided values. Default is ``None``.
    strip : bool, optional
        If ``True``, then once the user string is split, all values are stripped
        of whitespace on either side of the string before being converted to Python value
        by using Python's string ``strip()``.
        Default is ``True``.
    disregard_empty : bool, optional
        If ``True``, then once input string is split, all false evaluated values are
        disregarded. Default is ``True``.
    invalid_values : list, optional
        If provided, this list determines which values are invalid and if any are
        encountered, a ``ValidationError`` will be raised.
        Useful to blacklist specific values for being validated.
        If more complex logic is required, please consider using ``mapping``
        as a callable function.
    """
    default_error_messages = {
        'max_values': _('Ensure this value has at least {0} values (it has {1}).'),
        'min_values': _('Ensure this value has at most {0} values (it has {1}).'),
        'invalid_value': _('{0} is an invalid value.'),
        'invalid_values': _('{0} are invalid values.'),
    }

    def __init__(self,
                 delimiter=',',
                 separator=',',
                 mapping=None,
                 max_values=None,
                 min_values=None,
                 strip=True,
                 disregard_empty=True,
                 invalid_values=None,
                 *args, **kwargs):
        self.delimiter = delimiter
        self.separator = delimiter if not isinstance(delimiter, retype) else separator
        self.mapping = mapping or {}
        self.max_values = max_values
        self.min_values = min_values
        self.disregard_empty = disregard_empty
        self.invalid_values = set(invalid_values) if invalid_values else set()

        super(MultipleValuesCharField, self).__init__(*args, **kwargs)

        self.strip = strip

    def to_python(self, value):
        """
        Convert the input value to a list of values.

        Also does basic validation via the ``mapping`` especially
        when it is a callable and so can raise exceptions.

        Returns
        -------
        list
        """
        value = super(MultipleValuesCharField, self).to_python(value)
        if value in self.empty_values:
            return []

        if isinstance(self.delimiter, retype):
            values = self.delimiter.split(value)
        else:
            values = value.split(self.delimiter)

        if self.strip:
            values = [i.strip() for i in values]

        if self.disregard_empty:
            values = [i for i in values if i]

        if self.mapping:
            if not callable(self.mapping):
                values = [self.mapping.get(i, i) for i in values]
            else:
                invalid_values = []
                for i, v in enumerate(values):
                    try:
                        values[i] = self.mapping(v)
                    except Exception:
                        invalid_values.append(v)
                if invalid_values:
                    error_message = 'invalid_value' if len(invalid_values) == 1 else 'invalid_values'
                    raise forms.ValidationError(
                        self.error_messages[error_message].format(
                            ', '.join([six.text_type(i) for i in invalid_values])
                        )
                    )

        return values

    def validate(self, values):
        """
        Validate values

        Primarily this method validates:

        * that ``min_values`` and ``max_values`` are honored
        * that ``invalid_values`` were not supplied
        """
        if values in self.empty_values:
            if self.required:
                raise forms.ValidationError(
                    self.error_messages['required'], code='required'
                )
            else:
                return

        if self.min_values and len(values) < self.min_values:
            raise forms.ValidationError(
                self.error_messages['min_values'].format(len(values), self.min_values)
            )

        if self.max_values and len(values) > self.max_values:
            raise forms.ValidationError(
                self.error_messages['max_values'].format(len(values), self.max_values)
            )

        if self.invalid_values:
            invalid_values = set(values) & self.invalid_values
            if invalid_values:
                error_message = 'invalid_value' if len(invalid_values) == 1 else 'invalid_values'
                raise forms.ValidationError(
                    self.error_messages[error_message].format(
                        ', '.join(map(six.text_type, invalid_values))
                    )
                )

    def run_validators(self, values):
        """
        Run regular Django field validators for each of the values.
        """
        for v in values:
            super(MultipleValuesCharField, self).run_validators(v)

    def prepare_value(self, values):
        """
        Prepare values to be displayed in the UI.

        By default this method joins all the values by the
        ``separator`` if the values is a list or tuple.
        Otherwise, this method returns the value itself.
        Having this method does not require to define a
        custom widget for this form field.
        """
        if isinstance(values, (list, tuple)):
            return self.separator.join(values)
        return values



from __future__ import unicode_literals

from django.dispatch import Signal


sending_mail = Signal(providing_args=['message'])
"""
Signal sent when message is being sent.

Parameters
----------
message
    Email message which is being sent
"""
sent_mail = Signal(providing_args=['message'])
"""
Signal sent when message has successfully been sent

Parameters
----------
message
    Email message which was sent
"""
failed_mail = Signal(providing_args=['message', 'reason'])
"""
Signal sent when message sending has failed

Parameters
----------
message
    Email message which was not sent
reason : str
    Reason for why message was not sent
"""



from __future__ import print_function, unicode_literals

import mock
import pytest
from django import forms
from django.core.files.uploadedfile import UploadedFile
from django.core.validators import EMPTY_VALUES

from django_auxilium.forms import files
from django_auxilium.forms.files import TypedFileField


class TestTypedFileField(object):
    """
    Test the behaviour of the TypedFileField form field. Make sure that
    """
    good_extensions = ['txt', 'csv']
    bad_extensions = ['jpg', 'jpeg', 'png']
    good_types = ['text/plain', 'text/csv']
    bad_types = ['image/jpg', 'image/png']

    def test_empty_file(self):
        """
        Test that if empty file is given, None is returned.
        """
        field = TypedFileField(required=False)
        for v in EMPTY_VALUES:
            assert field.clean(v) is None

    def test_extensions(self):
        """
        Test that the extension validation is working properly
        """
        field = TypedFileField(required=False, ext_whitelist=self.good_extensions)

        for ext in self.good_extensions:
            name = 'somefooname.%s' % ext
            file = UploadedFile(name=name, size=1)
            assert field.clean(file) is file

        for ext in self.bad_extensions:
            name = 'somefooname.%s' % ext
            file = UploadedFile(name=name, size=1)
            with pytest.raises(forms.ValidationError):
                field.clean(file)

    def test_mimetypes(self):
        """
        Test that the mimetypes are validate correctly
        """
        field = TypedFileField(required=False, type_whitelist=self.good_types, use_magic=False)

        for t in self.good_types:
            name = 'somefooname'
            file = UploadedFile(name=name, size=1, content_type=t)
            assert field.clean(file) is file

        for t in self.bad_types:
            name = 'somefooname'
            file = UploadedFile(name=name, size=1, content_type=t)
            with pytest.raises(forms.ValidationError):
                field.clean(file)

    @mock.patch.object(files, 'get_content_type')
    def test_mimetypes_magic(self, mock_get_content_type):
        """
        Test that the mimetypes are validate correctly
        """

        def get_content_type(value):
            return value.content_type

        mock_get_content_type.side_effect = get_content_type

        field = TypedFileField(required=False, type_whitelist=self.good_types, use_magic=True)

        for t in self.good_types:
            name = 'somefooname'
            file = UploadedFile(name=name, size=1, content_type=t)
            assert field.clean(file) is file

        for t in self.bad_types:
            name = 'somefooname'
            file = UploadedFile(name=name, size=1, content_type=t)
            with pytest.raises(forms.ValidationError):
                field.clean(file)

    def test_no_mimetype(self):
        """
        Make sure ``ValidationError`` is raised if uploaded file has no mimetype
        """
        field = TypedFileField(required=False, type_whitelist=self.good_types, use_magic=False)

        for t in self.good_types:
            name = 'somefooname'
            file = UploadedFile(name=name, size=1, content_type=t)
            del file.content_type
            with pytest.raises(forms.ValidationError):
                field.clean(file)

    @mock.patch.object(files, 'get_content_type')
    def test_no_mimetype_magic(self, mock_get_content_type):
        """
        Make sure ``ValidationError`` is raised if uploaded file has no mimetype
        """
        mock_get_content_type.side_effect = ValueError

        field = TypedFileField(required=False, type_whitelist=self.good_types)

        for t in self.good_types:
            name = 'somefooname'
            file = UploadedFile(name=name, size=1, content_type=t)
            with pytest.raises(forms.ValidationError):
                field.clean(file)

    def test_both(self):
        """
        Test that both extensions and mimetypes are validated correctly both
        at the same time
        """
        field = TypedFileField(required=False,
                               ext_whitelist=self.good_extensions,
                               type_whitelist=self.good_types,
                               use_magic=False)

        for ext in self.good_extensions:
            name = 'somefooname.%s' % ext

            for t in self.good_types:
                file = UploadedFile(name=name, size=1, content_type=t)
                assert field.clean(file) is file

            for t in self.bad_types:
                file = UploadedFile(name=name, size=1, content_type=t)
                with pytest.raises(forms.ValidationError):
                    field.clean(file)

        for ext in self.bad_extensions:
            name = 'somefooname.%s' % ext

            for t in self.good_types:
                file = UploadedFile(name=name, size=1, content_type=t)
                with pytest.raises(forms.ValidationError):
                    field.clean(file)

            for t in self.bad_types:
                file = UploadedFile(name=name, size=1, content_type=t)
                with pytest.raises(forms.ValidationError):
                    field.clean(file)



from __future__ import print_function, unicode_literals
import abc
import types
from functools import partial, wraps

import six

from .decorators import HybridDecorator


class NotInCache(Exception):
    """
    Exception for when a value is not present in cache.

    Primary purpose of the exception is to normalize
    various exception types from different cache implementations.
    """


class BaseCache(six.with_metaclass(abc.ABCMeta, object)):
    """
    Base class for implementing cache implementations

    Parameters
    ----------
    parent : object
        A parent object where cache is stored
    attr : str
        Name of the attribute under which cache will be stored
        in the ``parent`` object
    """

    def __init__(self, parent, attr):
        self.parent = parent
        self.attr = attr

    def get(self, *args, **kwargs):
        """
        This method must be overwritten by subclasses which
        should return cache value

        ``args`` and ``kwargs`` are the parameters passed to
        the cached function and can be used by the method
        to extract the cache value from the ``parent`` object

        When cache value is not found, this method should raise
        :py:class:`NotInCache`
        """

    def set(self, value, *args, **kwargs):
        """
        This method must be overwritten by subclasses which
        should set the cache for the given parameters

        ``args`` and ``kwargs`` are the parameters passed to
        the cached function and can be used by the method
        to correctly store the cache value in the ``parent`` object
        """

    def delete(self, *args, **kwargs):
        """
        This method must be overwritten by subclasses which
        should delete cache value for the given parameters

        ``args`` and ``kwargs`` are the parameters passed to
        the cached function and can be used by the method
        to determine which cache value from the ``parent`` object
        to remove

        When cache value is not found, this method should raise
        :py:class:`NotInCache`
        """


class Caching(BaseCache):
    """
    Caching implementation which stores single cache value
    on the parent object

    The cache is simply stored on the given attribute.
    When the attribute is present on the ``parent`` object,
    that indicates that the cache was set and that value
    is used for cache. When not present, that means
    cache is missing and hence :py:class:`NotInCache` is raised
    in most use-cases.
    """

    def get(self, *args, **kwargs):
        """
        Get the cache value from the ``parent`` object

        Raises
        ------
        NotInCache
            When the cache is not set
        """
        try:
            return getattr(self.parent, self.attr)
        except AttributeError:
            raise NotInCache

    def set(self, value, *args, **kwargs):
        """
        Store the cache value on the ``parent`` object
        """
        setattr(self.parent, self.attr, value)
        return value

    def delete(self, *args, **kwargs):
        """
        Delete the cache value from the ``parent`` object

        Raises
        ------
        NotInCache
            When the cache is not set and so cannot be deleted
        """
        try:
            return self.parent.__dict__.pop(self.attr)
        except KeyError:
            raise NotInCache


class Memoizing(BaseCache):
    """
    Caching implementation which stores single cache value
    for a set of given parameters on the parent object
    hence is called memoization

    The cache is stored on the given attribute as a dictionary.
    Keys are string representations of the given parameters
    to the cached function and the values are their corresponding
    cache values. When the key is in the dictionary,
    that is used as cache value. Otherwise, in most cases
    :py:class:`NotInCache` is raised.
    """

    def _get_key(self, *args, **kwargs):
        return repr(args) + repr(sorted(kwargs.items()))

    def get(self, *args, **kwargs):
        """
        Get the cache value from the ``parent`` object
        by computing the key from the given parameters

        Raises
        ------
        NotInCache
            When the cache is not set
        """
        key = self._get_key(*args, **kwargs)
        try:
            return getattr(self.parent, self.attr)[key]
        except (AttributeError, TypeError, KeyError):
            raise NotInCache

    def set(self, value, *args, **kwargs):
        """
        Store the cache value on the ``parent`` object
        for the key as computed for the given parameters
        """
        key = self._get_key(*args, **kwargs)
        try:
            store = getattr(self.parent, self.attr)
        except AttributeError:
            store = {}
            setattr(self.parent, self.attr, store)
        store[key] = value
        return value

    def delete(self, *args, **kwargs):
        """
        Delete the cache value from the ``parent`` object
        for the key as computed by the given parameters

        Raises
        ------
        NotInCache
            When the cache is not set and so cannot be deleted
        """
        key = self._get_key(*args, **kwargs)
        try:
            return getattr(self.parent, self.attr).pop(key)
        except (AttributeError, TypeError, KeyError):
            raise NotInCache


class CacheDescriptor(object):
    """
    Cache descriptor to be used to add instance-level cache
    to class methods.

    .. note::
        This descriptor could be used independently (and there are
        even some examples below) however it is meant to be used
        along with :py:class:`CacheDecorator` which provides much
        cleaner API.

    Examples
    --------
    ::

        >>> def bar(self):
        ...     print('computing')
        ...     return 'bar'

        >>> class Foo(object):
        ...     foo = CacheDescriptor(bar)

        >>> f = Foo()
        >>> print(f.foo())
        computing
        bar
        >>> print(f.foo())
        bar
        >>> print(f.foo.pop())
        bar
        >>> print(f.foo())
        computing
        bar

    Parameters
    ----------
    method : function
        Callable which this descriptor is meant to wrap and cache
    cache_class : type, optional
        Caching implementation cache which should be used to
        apply caching. By default :py:attr:`.default_cache_class` is used.
    as_property : bool, optional
        Whether to implement the descriptor as a property.
        By default it is ``False``.

        .. warning::
            This option as ``True`` can only be used with some
            caching implementations such as :py:class:`Caching`.
            Other implementations do not suppose this.
    """
    cache_attribute_pattern = '{name}_cache_{hash}'
    """
    String pattern for constructing the cache attribute
    name under which cache will be stored on the instance.
    This attribute is meant to be changed in subclasses
    to customize the functionality.
    """
    default_cache_class = Caching
    """
    Cache implementation class which will be used
    for the caching.
    This attribute is meant to be changed in subclasses
    to customize the functionality.
    """

    def __init__(self, method, cache_class=None, as_property=False):
        self.method = method
        self.cache_attribute = self.cache_attribute_pattern.format(
            name=method.__name__,
            hash=abs(hash(method.__name__)),
        )
        self.cache_class = cache_class or self.default_cache_class
        self.as_property = as_property

    def get_cache(self, instance):
        """
        Helper method which given returns cache implementation instance
        for the given instance with given parameters
        """
        return self.cache_class(instance, self.cache_attribute)

    def getter(self, instance, *args, **kwargs):
        """
        Wrapper method around the decorator-wrapped callable
        (``method`` parameter) which returns cache value when available
        or otherwise computes and stores the value in cache by evaluating
        wrapped method
        """
        cache = self.get_cache(instance)

        try:
            return cache.get(*args, **kwargs)
        except NotInCache:
            return cache.set(self.method(instance, *args, **kwargs), *args, **kwargs)

    def pop(self, instance, *args, **kwargs):
        """
        Method for popping cache value corresponding to the given
        parameters from the instance as implemented by the
        caching implementation
        """
        cache = self.get_cache(instance)
        return cache.delete(*args, **kwargs)

    def _wrap(self, wrapping, proxy, instance):
        f = types.MethodType(proxy, instance)
        return wraps(wrapping)(partial(f))

    def __get__(self, instance, owner):
        if self.as_property:
            if instance is None:
                return self
            else:
                return self.getter(instance)

        else:
            if instance is None:
                return self.method
            else:
                f = self._wrap(self.method, self.getter, instance)
                f.pop = self._wrap(self.__class__.pop, self.pop, instance)
                return f

    def __set__(self, instance, value):
        if self.as_property:
            cache = self.get_cache(instance)
            # no args and kwargs since this is only for case when cache is used as property
            return cache.set(value)

        # required to be overwritten for pypy
        # see https://bitbucket.org/pypy/pypy/issues/2033/attributeerror-object-attribute-is-read
        # even though ticket is resolved its still affecting pypy3
        raise AttributeError

    def __delete__(self, instance):
        if self.as_property:
            self.pop(instance)
        else:
            raise AttributeError


class MemoizeDescriptor(CacheDescriptor):
    """
    Cache descriptor to be used to add instance-level cache
    to class methods which considers method parameters
    when caching values.

    In other words, this descriptor caches method results
    per given parameters.

    .. note::
        This descriptor could be used independently (and there are
        even some examples below) however it is meant to be used
        along with :py:class:`MemoizeDecorator` which provides much
        cleaner API.

    Examples
    --------
    ::

        >>> def bar(self, x):
        ...     print('computing for', x)
        ...     return x + 'bar'

        >>> class Foo(object):
        ...     foo = MemoizeDescriptor(bar)

        >>> f = Foo()
        >>> print(f.foo('foo'))
        computing for foo
        foobar
        >>> print(f.foo('awesome'))
        computing for awesome
        awesomebar
        >>> print(f.foo('foo'))
        foobar
        >>> print(f.foo.pop('foo'))
        foobar
        >>> print(f.foo('foo'))
        computing for foo
        foobar
        >>> print(f.foo('awesome'))
        awesomebar
    """
    cache_attribute_pattern = '{name}_memoize_{hash}'
    """
    String pattern for constructing the cache attribute
    name under which cache will be stored on the instance.
    This attribute is meant to be changed in subclasses
    to customize the functionality.
    """
    default_cache_class = Memoizing
    """
    Cache implementation class which will be used
    for the caching.
    This attribute is meant to be changed in subclasses
    to customize the functionality.
    """


class BaseCacheDecorator(HybridDecorator):
    """
    Base decorator for caching callables so that they only execute once
    and all subsequent calls return cached value

    This is very useful for expensive functions
    """
    cache_descriptor_class = None
    """
    Descriptor class to be used when caching is applied to class methods.
    This attribute is meant to be changed in subclasses.
    """
    cache_class = Caching
    """
    Caching implementation to use when wrapping standalone functions.
    This attribute is meant to be changed in subclasses.
    """

    def get_cache_descriptor(self):
        """
        Hook for instantiating cache descriptor class
        """
        return self.cache_descriptor_class(self.to_wrap)

    def get_wrapped_object(self):
        """
        Main method for wrapping the given object.

        This method has separate code paths for the following scenarios:

        :class method:
            When wrapping class method descriptor is returned
            as created by :py:meth:`get_cache_descriptor`.
            That descriptor is then responsible for maintaining
            cache state for the class instance.
        :function:
            When wrapping standalone function, this method
            returns a wrapping function which caches the
            value on this decorator instance.
            That means that the cache becomes global
            since functions in Python are singletons.
        """
        to_wrap = super(BaseCacheDecorator, self).get_wrapped_object()

        if self.in_class:
            return self.get_cache_descriptor()

        else:
            self.cache = self.cache_class(self, 'cached_value')

            def wrapper(*args, **kwargs):
                try:
                    return self.cache.get(*args, **kwargs)
                except NotInCache:
                    return self.cache.set(to_wrap(*args, **kwargs), *args, **kwargs)

            wrapper.pop = self.pop
            wrapper.decorator = self
            return wrapper

    def pop(self, *args, **kwargs):
        """
        Method for popping cache value corresponding to the given parameters
        """
        return self.cache.delete(*args, **kwargs)


class CacheDecorator(BaseCacheDecorator):
    """
    Decorator for caching callables so that they only execute once
    and all subsequent calls return cached value.

    .. note::
        This caching decorator does not account function parameters
        to cache values. If you need to cache different calls
        per given parameters, please look at :py:class:`MemoizeDecorator`

    Examples
    --------

    When caching standalone functions::

        >>> @CacheDecorator.as_decorator()
        ... def compute():
        ...     print('computing here')
        ...     return 'foo'

        >>> print(compute())
        computing here
        foo
        >>> print(compute())
        foo

        >>> print(compute.pop())
        foo

        >>> print(compute())
        computing here
        foo

    When caching class methods::

        >>> class Foo(object):
        ...     @CacheDecorator.as_decorator()
        ...     def foo(self):
        ...         print('computing here')
        ...         return 'foo'

        >>> f = Foo()
        >>> print(f.foo())
        computing here
        foo
        >>> print(f.foo())
        foo

        >>> print(f.foo.pop())
        foo

        >>> print(f.foo())
        computing here
        foo

    Parameters
    ----------
    as_property : bool
        Boolean whether to create a property descriptor
        when using it on a class method

        .. warning::
            This is only meant to be used when the wrapping
            method does not accept any parameters since there
            is no way in Python to pass parameters to properties
    """
    cache_descriptor_class = CacheDescriptor
    """
    Descriptor class to be used when caching is applied to class methods
    """
    cache_class = Caching
    """
    Caching implementation to use when wrapping standalone functions.
    """

    def __init__(self, as_property=False, *args, **kwargs):
        self.as_property = as_property
        super(CacheDecorator, self).__init__(*args, **kwargs)

    def get_cache_descriptor(self):
        """
        Custom implementation for getting the cache descriptor
        which allows to use ``as_property`` parameter
        """
        return self.cache_descriptor_class(
            self.to_wrap, as_property=self.as_property,
        )


class MemoizeDecorator(BaseCacheDecorator):
    """
    Decorator for memoizing functions so that they only execute once
    and all subsequent calls with same parameters

    This is very useful for expensive functions which accept parameters

    Examples
    --------

    When caching standalone functions::

        >>> @MemoizeDecorator.as_decorator()
        ... def compute(x):
        ...     print('computing for', x)
        ...     return x + 'foo'

        >>> print(compute('bar'))
        computing for bar
        barfoo
        >>> print(compute('awesome'))
        computing for awesome
        awesomefoo
        >>> print(compute('bar'))
        barfoo

        >>> print(compute.pop('bar'))
        barfoo

        >>> print(compute('bar'))
        computing for bar
        barfoo
        >>> print(compute('awesome'))
        awesomefoo

    When caching class methods::

        >>> class Foo(object):
        ...     @MemoizeDecorator.as_decorator()
        ...     def foo(self, x):
        ...         print('computing for', x)
        ...         return x + 'foo'

        >>> f = Foo()
        >>> print(f.foo('bar'))
        computing for bar
        barfoo
        >>> print(f.foo('awesome'))
        computing for awesome
        awesomefoo

        >>> print(f.foo('bar'))
        barfoo
        >>> print(f.foo('awesome'))
        awesomefoo

        >>> print(f.foo.pop('bar'))
        barfoo

        >>> print(f.foo('bar'))
        computing for bar
        barfoo
        >>> print(f.foo('awesome'))
        awesomefoo
    """
    cache_descriptor_class = MemoizeDescriptor
    """
    Descriptor class to be used when caching is applied to class methods
    """
    cache_class = Memoizing
    """
    Caching implementation to use when wrapping standalone functions.
    """


cache = CacheDecorator.as_decorator()
memoize = MemoizeDecorator.as_decorator()
cache_property = CacheDecorator.as_decorator(as_property=True)
"""
Shortcut for :py:data:`cache` which automatically creates
properties when used on class methods.

These are equivalent::

    @cache(as_property=True)
    def foo(self): pass

    @cache_property
    def bar(self): pass
"""
cache_method = CacheDecorator.as_decorator(is_method=True)
"""
Shortcut for :py:data:`cache` which automatically indicates
that the cached object is a method.

These are equivalent::

    class Foo(object):
        @cache(is_method=True)
        def foo(self): pass

    class Foo(object):
        @cache_method
        def bar(self): pass
"""



from __future__ import print_function, unicode_literals
import re

from django import forms
from django.utils.translation import ugettext_lazy as _


class ColorCharField(forms.CharField):
    """
    Custom django field which is able to validate colors as input.

    Parameters
    ----------
    type : str
        The type of the color format. The supported formats are

        :hex:
            The hexadecimal notation of the color. Examples is::

                ff0000
                #00ff00

            This type also allows to specify these additional parameters:

            Parameters
            ----------

            hash_required : bool
                If the ``#`` is required in the value
                (e.g. ``'#ff0000'``)
            hash_output : bool
                If the ``#`` is required in the output
                validated value in ``to_python``


    Raises
    ------
    ValueError
        If unsupported ``type`` color type is provided
    """

    default_error_messages = {
        'invalid': _('Invalid color value.'),
    }

    def __init__(self, *args, **kwargs):
        self.color_type = kwargs.pop('type', 'hex')

        if self.color_type == 'hex':
            kwargs['min_length'] = 6
            kwargs['max_length'] = 7
            for p in ['hash_required', 'hash_output']:
                setattr(self, p, kwargs.pop(p, None))

        else:
            raise ValueError('Unsupported color type.')

        super(ColorCharField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        """
        Convert to a validated string color value.
        """
        value = super(ColorCharField, self).to_python(value)
        if value in self.empty_values:
            return ''

        if self.color_type == 'hex':
            value = value.upper()

            # calculate the validation regex
            if self.hash_required:
                r = re.compile(r'^#[0-9A-F]{6}$')
            else:
                r = re.compile(r'^#?[0-9A-F]{6}$')

            # validate the format of the string
            if not r.findall(value):
                raise forms.ValidationError(self.error_messages['invalid'])

            # remove hash if exists
            if value[0] == '#':
                value = value[1:]

            # add hash only if required for output
            if self.hash_output:
                value = '#' + value

        return value



from __future__ import print_function, unicode_literals
import inspect
from collections import namedtuple

import six
from django.db import models
from django.dispatch.dispatcher import Signal

from django_auxilium.utils.functools import Decorator, cache


FieldSpec = namedtuple('FieldSpec', ['name', 'field'])
"""
Field specification which is used by file descriptors to reference
the field name and field object itself during decorator operations.
"""


class FileFieldAutoDelete(Decorator):
    """
    Model decorator which automatically setups all the necessary signals to automatically
    remove files associated with given fields when model instance is removed.

    Starting with Django 1.3 when model records are deleted via querysets (e.g.
    ``model.objects.filter(...).delete()``), the model's ``delete()`` method is no longer
    called. As a consequence, if the model has any file fields, those files will not
    be removed even if it is specified in the ``delete()`` method. The new way to remove
    files associated with a model is to use Django signals framework, or more specifically
    the ``post_delete`` signal. This decorator automatically connects the ``post_delete``
    signal which will remove the file associated with the specified file field.

    More about this can be found at Django 1.3
    `release notes <https://docs.djangoproject.com/en/dev/releases/1.3/#deleting-a-model-doesn-t-delete-associated-files>`_.

    Examples
    --------

    ::

        >>> @file_field_auto_delete  # equivalent to @file_field_auto_delete('*')
        ... class FooModel(models.Model):
        ...     file = models.FileField(upload_to='foo')

        >>> @file_field_auto_delete('file')
        ... class FooModel(models.Model):
        ...     file = models.FileField(upload_to='foo')

    Parameters
    ----------
    fields : str, list
        Either a single file field which should be removed on delete
        or a list of fields. Also ``'*'`` can be used in order to delete all
        ``FileField`` on the model.
    signal : Signal, optional
        Djagno signal class which should be used to attach the receiver signal handler to.
        By default is ``post_delete`` signal.
    signal_name_pattern : str, optional
        The name given to the signal function which will automatically remove the file.
        This can be a string formatted string. Into it, two parameters will be passed
        which are ``model`` and ``field``. ``model`` is a model class (not instance) to
        which the signal is being connected to and ``field`` is a name of the file field
        (or ``'_'`` delimited field names if multiple fields are given).
        The default pattern is ``post_delete_{model.__name__}_delete_{field}`` which
        results in patterns like ``'post_delete_Model_delete_file_field'``. The reason
        why this pattern might be useful is because it can be used to disconnect the
        signal receiver at a later time.

    Attributes
    ----------
    field_names : list
        List of normalized field names which should be removed on change
    fields : list
        List of validated :py:class:`.FieldSpec` which signal
        will use to remove changed files
    signal_name_pattern : str
        Same as the ``signal_name_pattern`` parameter
    """

    def __init__(self, fields='*', signal=None, signal_name_pattern=None):
        self.fields = []
        self.field_names = fields
        self.signal = signal or models.signals.post_delete
        self.signal_name_pattern = signal_name_pattern or 'post_delete_{model.__name__}_delete_{field}'

    def get_wrapped_object(self):
        """
        Return the given model

        Since signals are connected externally of a model class,
        the model class is not modified.

        See Also
        --------
        validate_model
        get_field_names
        validate_fields
        connect_signal_function
        """
        self.validate_model()
        self.field_names = self.get_field_names()
        self.validate_fields()
        self.fields = self.get_fields()
        self.connect_signal_function()

        return self.to_wrap

    def get_fields(self):
        """
        Get list of :py:class:`.FieldSpec` for all the fields
        to be removed
        """
        return [
            FieldSpec(name, self.get_field_by_name(name))
            for name in self.field_names
        ]

    def get_field_names(self):
        """
        Get list of field names to be removed in the created signal

        This method is primarily used to normalize the list of
        fields in case it is provided as a string, list or a
        wildcard ``'*'``.

        Returns
        -------
        list
            List of field names which should be removed
            in the created signal
        """
        if isinstance(self.field_names, (list, tuple)):
            return self.field_names

        elif self.field_names == '*':
            return [
                i.name for i in
                self.to_wrap._meta.fields
                if isinstance(i, models.FileField)
            ]

        else:
            return [self.field_names]

    def get_field_by_name(self, name):
        """
        Get a Django model field for the given field name
        from the decorated model

        Parameters
        ----------
        name : str
            Name of the field to get

        Returns
        -------
        Field
            Django model field instance
        """
        return next(iter(filter(
            lambda f: f.name == name,
            self.to_wrap._meta.fields
        )), None)

    def validate_model(self):
        """
        Validate the input to the decorator which is suppose to be a model

        Make sure the given class is a subclass of Django's ``Model`` class.

        Raises
        ------
        TypeError
            If ``to_wrap`` is either not a class or not a Django's
            ``model.Model`` class.
        """
        if not inspect.isclass(self.to_wrap):
            raise TypeError('This decorator can only be applied to classes')
        if not issubclass(self.to_wrap, models.Model):
            raise TypeError('This decorator can only be applied to Django models')

    def validate_fields(self):
        """
        Validate all the fields on the model

        This method is expected to be called after :py:meth:`validate_model`
        and it assumes that ``self.to_wrap`` is a valid Django model.
        This method validates that the given fields are present in the model
        and that they are a subclass of Django's ``FileField`` field class.

        See Also
        --------
        validate_field
        """
        for field in self.field_names:
            self.validate_field(field)

    def validate_field(self, name):
        """
        Validate a single field on the model

        This method is expected to be called after :py:meth:`validate_model`
        and it assumes that ``self.to_wrap`` is a valid Django model.
        It also validates that the given field name is present in the model
        and that it subclasses Django's ``FileField`` field class.

        Raises
        ------
        AttributeError
            When the field cannot be found on the model
        TypeError
            When the field is not a ``FileField``
        """
        field = self.get_field_by_name(name)

        if field is None:
            raise AttributeError('Field "{0}" cannot be found'.format(name))

        if not isinstance(field, models.FileField):
            m = 'Field "{0}" must be a subclass of Django ``FileField``'
            raise TypeError(m.format(name))

    def get_signal_name(self):
        """
        Using the :py:attr:`signal_name_pattern` pattern,
        return the formatted signal name.

        Returns
        -------
        name : str
            The name of the signal
        """
        return self.signal_name_pattern.format(
            model=self.to_wrap,
            field='_'.join(i.name for i in self.fields),
        )

    @cache
    def get_signal_function(self):
        """
        Get the actual function which will be connected to the Django's signal which
        conforms to the ``post_delete`` signal signature::

            def receiver(sender, instance, *args, **kwargs): pass
        """

        def remove(sender, instance, *args, **kwargs):
            """
            Automatically remove the file field when model instance is deleted.
            """
            for name, field in self.fields:
                value = getattr(instance, name, None)
                if value:
                    method = getattr(value, 'delete', None)
                    if method and callable(method):
                        method(save=False)

        return remove

    def connect_signal_function(self):
        """
        Connect the signal as returned by :py:meth:`get_signal_function`
        into the Django's signal's framework.
        """
        signal_name = self.get_signal_name()
        signal_function = self.get_signal_function()
        self.signal.connect(
            signal_function,
            sender=self.to_wrap,
            weak=False,
            dispatch_uid=signal_name
        )


class FileFieldAutoChangeDelete(FileFieldAutoDelete):
    """
    Model decorator which automatically setups all the necessary signals to automatically
    remove files associated with given fields when they change on model save.

    Unlike :py:class:`FileFieldAutoDelete` which removed file fields on
    model deletion, this decorator removes files if they are changed.
    The premise is that if the file field is editable if it is changed
    either in Django Admin or in custom logic, unless custom logic
    is applied, by default Django would not remove the old file
    hence over time many orphan files can accumulate on the server.
    Locally it might not be a big deal however when external (especially paid)
    storage backend is used, extra files can make a real difference over time.
    This decorator automatically handles that and removes old files when
    they are changed.

    In order to do that however model has to track when files change
    on the model and Django models do not do that out of the box.
    You can either implement that functionality yourself on the model
    by implementing the following methods:

    * ``is_dirty()`` - should return a ``bool`` if any of the model
      fields have changed. Usually this will require to track model
      fields during ``__init__`` which will be used to initialize
      model instance when querying model from db and then comparing
      those values with the new values when this method is called.
    * ``get_dirty_fields()`` - should return a ``dict`` where the
      keys are the field names and the values are old field values.

    Alternatively you can use a third-party package which implements
    that API. This decorator is tested with
    `django-dirtyfields <https://pypi.python.org/pypi/django-dirtyfields/>`_.

    .. warning::
        As mentioned above, this decorator **requires** ``is_dirty()``
        and ``get_dirty_fields()`` to be implemented on the model.
        Even though we recommend to use ``django-dirtyfields`` for that
        functionality, this library does not ship with ``django-dirtyfields``
        pre-installed and you will need to install it independently.

    .. note::
        This decorator does not remove files on delete as well.
        If you would like to both remove files on both delete and change
        you will need to apply both :py:class:`FileFieldAutoDelete` and this
        decorator at the same time.

    Examples
    --------

    ::

        >>> from dirtyfields import DirtyFieldsMixin

        >>> @file_field_auto_change_delete  # equivalent to @file_field_auto_change_delete('*')
        ... class FooModel(DirtyFieldsMixin, models.Model):
        ...     file = models.FileField(upload_to='foo')

        >>> @file_field_auto_change_delete('file')
        ... class FooModel(DirtyFieldsMixin, models.Model):
        ...     file = models.FileField(upload_to='foo')

        >>> # remote both on delete and change
        >>> @file_field_auto_delete
        ... @file_field_auto_change_delete
        ... class FooModel(DirtyFieldsMixin, models.Model):
        ...     file = models.FileField(upload_to='foo')

    Parameters
    ----------
    fields : str, list
        Same as :py:class:`FileFieldAutoDelete` ``fields`` parameter
    signal : Signal, optional
        Same as :py:class:`FileFieldAutoDelete` ``signal`` parameter.
        By default is ``post_save`` signal.
    signal_name_pattern : str, optional
        The name given to the signal function which will automatically remove the file.
        This can be a string formatted string. Into it, two parameters will be passed
        which are ``model`` and ``field``. ``model`` is a model class (not instance) to
        which the signal is being connected to and ``field`` is a name of the file field
        (or ``'_'`` delimited field names if multiple fields are given).
        The default pattern is ``post_save_{model.__name__}_delete_{field}`` which
        results in patterns like ``'post_save_Model_delete_file_field'``. The reason
        why this pattern might be useful is because it can be used to disconnect the
        signal receiver at a later time.
    """

    def __init__(self, *args, **kwargs):
        kwargs['signal'] = kwargs.pop('signal', models.signals.post_save)
        kwargs['signal_name_pattern'] = kwargs.pop(
            'signal_name_pattern',
            'post_save_{model.__name__}_changedelete_{field}'
        )
        super(FileFieldAutoChangeDelete, self).__init__(*args, **kwargs)

    def validate_model(self):
        """
        Validate that the model is a valid Django model and that
        it implements necessary API in order to track that model fields
        have changed before saving.
        """
        super(FileFieldAutoChangeDelete, self).validate_model()

        if not all([callable(getattr(self.to_wrap, 'is_dirty', None)),
                    callable(getattr(self.to_wrap, 'get_dirty_fields', None))]):
            raise ValueError(
                'Given model must implement `is_dirty` and '
                '`get_dirty_fields` methods. '
                'You can install "django-dirtyfields" which implements '
                'checking if fields changed before save.'
            )

    @cache
    def get_signal_function(self):
        """
        Get the actual function which will be connected to the Django's signal which
        conforms to the ``post_save`` signal signature::

            def receiver(sender, instance, *args, **kwargs): pass
        """
        def autoremove(sender, instance, *args, **kwargs):
            if instance.is_dirty():
                dirty_fields = instance.get_dirty_fields()
                for name, field in self.fields:
                    # while removing file Django also deletes
                    # reference to it on the model
                    # hence new file will be lost
                    # so we save it and then restore
                    # after removing old file
                    new = getattr(instance, name)
                    old = dirty_fields.get(name, None)
                    if old:
                        # the way Django pickles FileField, it does not
                        # preserve all instance attributes
                        # since they are normally reset by the
                        # FileFieldDescriptor hence we need to manually
                        # restore them here because get_dirty_fields()
                        # probably returns copies of the fields
                        # since it needs to compare them to determine
                        # what is dirty and what is not
                        old.instance = instance
                        old.field = field
                        old.storage = field.storage
                        old.delete(save=False)
                        setattr(instance, name, new)

        return autoremove


class AutoSignals(Decorator):
    """
    Decorator for automatically attaching signals to the model class.

    Usual Django convention is to include all signals in ``signals.py``.
    That works really well in most cases however here are *potential*
    disadvantages:

    * it couples logic to a model class in different Python module
    * signals need to be imported somewhere in app initialization process

    These issues can be resolved by moving the signal receivers to
    ``models.py`` however some issues with that are:

    * signals need to be defined after model definitions
    * if there are many models, it could be hard to find associated signals

    This decorator attempts to solve that by moving the signal receiver
    definitions to the model itself which are retrieved via
    class method ``get_signals``. It allows to tightly couple model
    signal receivers within the model which in some cases might be
    better compared to using ``signals.py`` module.

    ``get_signals`` expected to return a list of signals which should
    be attached to the model class. That could either be achieved
    by returning either receiver handler callables or signal dict
    parameters. Here is a description of what each of them can do:

    :Callables:
        An example::

            >>> @auto_signals
            ... class MyModel(models.Model):
            ...     @staticmethod
            ...     def do_pre_save(sender, instance, *args, **kwargs):
            ...         print('triggered pre_save')
            ...     @staticmethod
            ...     def do_post_save(sender, instance, *args, **kwargs):
            ...         print('triggered post_save')
            ...     @classmethod
            ...     def get_signals(cls):
            ...         return [cls.do_pre_save, cls.do_post_save]

        When ``get_signals`` returns a list of signal receivers,
        each receiver function name should contain the signal name
        to which it needs to connect to. By default this decorator
        supports all Django model signals. If different signals need
        to be supported, please use ``signal_pool`` parameter.
    :Dictionary:
        An example::

            >>> @auto_signals
            ... class MyModel(models.Model):
            ...     @classmethod
            ...     def get_signals(cls):
            ...         def so_something(sender, instance, *args, **kwargs):
            ...             print('doing something')
            ...         def do_something_else_post_save(sender, instance, *args, **kwargs):
            ...             print('doing something else')
            ...         return [
            ...             {
            ...                 'receiver': so_something,
            ...                 'weak': False,
            ...                 'dispatch_uid': 'my_signal_uuid',
            ...                 'signal': 'pre_save',
            ...             },
            ...             {
            ...                 'receiver': do_something_else_post_save,
            ...                 'weak': False,
            ...                 'dispatch_uid': 'my_other_signal_uuid',
            ...             },
            ...         ]

        When ``get_signals`` returns a list of signal dictionaries,
        each ``dict`` should be parameters which will be passed to
        Django's ``Signal.connect``. The only exception is the optional
        ``signal`` key which when given should either be a name of the
        supported signal as per ``signal_pool`` or actual signal object.
        When not provided, this decorator will figure out the signal
        from the receiver's name, same as when ``get_signals`` returns
        a list of callables as described above.

    Parameters
    ----------
    getter : str, optional
        Name of the class method on the decorated model which
        should return signals to be connected
    signal_pool : dict, optional
        Dictionary of supported signals. Keys should be signal
        names and values should be actual signal objects.
        By default all model signals are supported.
    """

    def __init__(self, getter='get_signals', signal_pool=None):
        self.getter = getter
        if signal_pool is None:
            signal_pool = {
                k: v for k, v in vars(models.signals).items()
                if isinstance(v, Signal)
            }
        self.signal_pool = signal_pool

    def validate_model(self):
        """
        Validate that the decorated object is a valid Django model
        class and that it implements signal getter.
        """
        if not inspect.isclass(self.to_wrap):
            raise TypeError('This decorator can only be applied to classes')
        if not issubclass(self.to_wrap, models.Model):
            raise TypeError('Decorator can only be applied to Django models')
        if not hasattr(self.to_wrap, self.getter):
            raise AttributeError(
                'Provided model must implement "{0}" method'
                ''.format(self.getter)
            )
        if not callable(getattr(self.to_wrap, self.getter)):
            raise TypeError(
                'Provided model must implement "{0}" method'
                ''.format(self.getter)
            )

    def get_wrapped_object(self):
        """
        Return the given model

        Since signals are connected externally of a model class,
        the model class is not modified.

        See Also
        --------
        validate_model
        connect_signals
        """
        self.validate_model()
        self.connect_signals()
        return self.to_wrap

    def connect_signals(self):
        """
        Connect all signals as returned by the signal getter
        on the decorated model

        See Also
        --------
        connect_signal
        """
        signals = getattr(self.to_wrap, self.getter)()
        if not isinstance(signals, (list, tuple)):
            signals = [signals]
        for signal in signals:
            self.connect_signal(signal)

    def connect_signal(self, signal):
        """
        Connect individual signal to the decorated model

        Parameters
        ----------
        signal : dict, function
            Either a dict or a function of the receiver signal handler.
            For more information about how each of those are handled,
            please :py:class:`AutoSignals` description.
        """
        kwargs = {
            'sender': self.to_wrap,
        }

        if isinstance(signal, dict):
            kwargs.update(signal)
        elif callable(signal):
            kwargs.update({
                'receiver': signal,
            })
        else:
            raise TypeError(
                'Signals should either be given as a `dict` or callables. '
                '{0} was given.'
                ''.format(type(signal).__name__)
            )

        if 'receiver' not in kwargs:
            raise ValueError(
                'Missing signal receiver callable. '
                'Please either provide a "receiver" key in the signal dict '
                'or use a callable instead of a dict.'
            )

        # if signal to connect to was not provided explicitly
        # try to determine it from the callable name
        if 'signal' not in kwargs:
            for s in self.signal_pool:
                if s in kwargs['receiver'].__name__:
                    kwargs['signal'] = s
                    break

            # the callable name does not suggest which signal to use
            if 'signal' not in kwargs:
                raise ValueError(
                    'Signal to connect to was not provided.'
                )

        # get the signal callable
        if isinstance(kwargs['signal'], six.string_types):
            try:
                kwargs['signal'] = self.signal_pool[kwargs['signal']]
            except KeyError:
                raise ValueError(
                    'Invalid signal "{}" was provided. '
                    'Please either explicitly specify Django signal '
                    'or provide its name in the receiver callable name.'
                    ''.format(kwargs['signal'])
                )

        kwargs.pop('signal').connect(**kwargs)


file_field_auto_delete = FileFieldAutoDelete.as_decorator()
file_field_auto_change_delete = FileFieldAutoChangeDelete.as_decorator()
auto_signals = AutoSignals.as_decorator()



from __future__ import print_function, unicode_literals

from django.core.mail.backends.smtp import EmailBackend

from .base import BaseCallbackEmailBackendMixin


class CallbackEmailBackend(BaseCallbackEmailBackendMixin, EmailBackend):
    """
    Custom SMPT email backend which sends signals on progress events
    of sending messages.
    """

    def send_messages(self, email_messages):
        """
        Custom implementation of sending messages which also
        calls mail signals.

        Unfortunately vanilla Django implementation does not
        provide enough hooks for when messages are sent
        hence this method needs to duplicate the whole method.
        """
        if not email_messages:
            return
        with self._lock:
            new_conn_created = self.open()
            if not self.connection:
                # We failed silently on open().
                # Trying to send would be pointless.
                for message in email_messages:
                    self.failed(message, 'Could not establish connection')
                return
            num_sent = 0
            for message in email_messages:
                self.sending(message)
                sent = self._send(message)
                if sent:
                    self.sent(message)
                    num_sent += 1
                else:
                    self.failed(message, 'Could not send email over SMTP')
            if new_conn_created:
                self.close()
        return num_sent



# Bare ``settings.py`` for running tests for django_auxilium

DEBUG = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'django_auxilium.sqlite'
    }
}

INSTALLED_APPS = [
    'django_auxilium',
]

MIDDLEWARE_CLASSES = []

STATIC_URL = '/static/'
SECRET_KEY = 'foo'



from __future__ import division, print_function, unicode_literals
from collections import namedtuple


class Range(namedtuple('rangenamedtuple', ['first_column', 'first_row', 'second_column', 'second_row'])):
    """
    Helper data-structure for storing excel-type ranges.

    Also has some useful range-related properties
    """

    @property
    def columns(self):
        """
        Gets the number of columns in a given range.

        Returns
        -------
        columns : int
            The number of columns
        """
        if self.first_column and self.second_column:
            return self.second_column - self.first_column + 1
        return 0

    @property
    def rows(self):
        """
        Gets the number of columns in a given range.

        Returns
        -------
        columns : int
            The number of columns
        """
        if self.first_row and self.second_row:
            return self.second_row - self.first_row + 1
        return 0


class AlphabeticNumbers(object):
    """
    Utility class for dealing with numbers in alphabetical form. Useful for converting
    Excel column notation into numbers and vise verse.
    """

    @staticmethod
    def int_from_str(string):
        """
        Convert alphabetical number into actual integer.

        Parameters
        ----------
        string : str
            Number in alphabetical form (e.g. AB = 28)

        Returns
        -------
        number : int
            Converted integer value
        """
        string = string.upper()[::-1]
        number = 0
        for i, c in enumerate(string):
            number += (ord(c) - 64) * (26 ** i)
        return number

    @staticmethod
    def str_from_int(number):
        """
        Convert integer into its alphabetical representation.

        Parameters
        ----------
        number : int
            Number to be converted

        Returns
        -------
        string : str
            The alphabetical representation of the number

        Raises
        ------
        ValueError
            If the input number is not valid
        """

        if number < 1:
            raise ValueError('Invalid Number')

        string = ''

        # recursive method for education purposes
        # however regular loop is used due to better performance
        #
        # if number < 27:
        #     return chr(number + 64)
        # else:
        #     remainder = number % 26
        #     if not remainder:
        #         return AlphabeticNumbers.str_from_int(int(number / 26) - 1) + 'Z'
        #     else:
        #         return AlphabeticNumbers.str_from_int(int(number / 26)) + chr(remainder + 64)

        while True:
            remainder = number % 26
            if not remainder:
                string += 'Z'
                number = number // 26 - 1
            else:
                string += chr(remainder + 64)
                number //= 26
            if number < 1:
                break

        return string[::-1]



from __future__ import print_function, unicode_literals
import pathlib

from django import forms
from django.utils.translation import ugettext_lazy as _

from django_auxilium.utils.content_type import get_content_type


class TypedFileField(forms.FileField):
    """
    Typed FileField which allows to make sure the uploaded file has
    the appropriate type.

    File type can be verified either by extension and/or mimetype.

    This field accepts all the parameters as FileField, however in addition
    it accepts some additional parameters as documented below.

    Examples
    --------

    ::

        TypedFileField(ext_whitelist=['jpg', 'jpeg'],
                       type_whitelist=['image/jpeg'])

    .. warning::
        If ``use_magic`` is used, please make sure that ``python-magic``
        is installed. This library does not require it by default.

    Parameters
    ----------
    ext_whitelist : list, optional
        List of allowed file extensions. Note that each extension
        should emit the first period. For example for filename
        ``'example.jpg'``, the allowed extension should be ``'jpg'``.
    type_whitelist : list, optional
        List of allowed file mimetypes.
    use_magic : bool, optional
        If ``type_whitelist`` is specified, this boolean determines
        whether to use ``python-magic`` (based of top of ``libmagic``)
        to determine the mimetypes of files instead of relying
        on Django's ``UploadedFile.content_type``.
        Django does not take any real care to determine any accurately
        the mimetype so it is recommended to leave this parameter
        as ``True`` which is the default value.
    """

    default_error_messages = {
        'extension': _('File extension is not supported.'),
        'mimetype': _('File mimetype is not supported.')
    }

    def __init__(self, *args, **kwargs):
        """
        Init methods extends the default FileField init however adds
        support for extra parameters
        """
        ext_whitelist = kwargs.pop('ext_whitelist', [])
        type_whitelist = kwargs.pop('type_whitelist', [])

        self.ext_whitelist = [i.lower() for i in ext_whitelist]
        self.type_whitelist = [i.lower() for i in type_whitelist]
        self.use_magic = kwargs.pop('use_magic', True)

        super(TypedFileField, self).__init__(*args, **kwargs)

    def validate(self, value):
        """
        Validate that the file is of supported extension and/or type.
        """
        super(TypedFileField, self).validate(value)
        if value in self.empty_values:
            return None

        # make sure the extension is correct
        ext = pathlib.Path(value.name).name.split('.', 1)[-1]
        if self.ext_whitelist and ext not in self.ext_whitelist:
            raise forms.ValidationError(self.error_messages['extension'])

        # getting mimetype of a file could use some resources
        # so if we don't need to check mimetype, we can return faster
        if not self.type_whitelist:
            return value

        # get the mimetype from the UploadedFile
        if self.use_magic:
            try:
                mimetype = get_content_type(value)
            except Exception:
                raise forms.ValidationError(self.error_messages['invalid'])

        else:
            try:
                mimetype = value.content_type
            except AttributeError:
                raise forms.ValidationError(self.error_messages['invalid'])

        # make sure the mimetype is correct
        if mimetype not in self.type_whitelist:
            raise forms.ValidationError(self.error_messages['mimetype'])

        return value



"""
This package provides provides custom email backends
which use Django signals framework to send signals
when various mail sending progress events occur.

You can refer to :py:mod:`.signals` for a full
list of supported signals.

In order to enable this functionality, you need to
change your Django email backend::

    EMAIL_BACKEND = 'django_auxilium.callback_mail.filebased.EmailBackend'

Examples
--------

::

    from django.dispatch import receiver
    from django_auxilium.callback_mail.signals import sent_mail

    @receiver(sent_mail)
    def my_callback(sender, message):
        print("Message sent!")
"""



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import six

from .functools import Decorator


class TextOnion(Decorator):
    """
    Simple text onion decorator.

    This decorator ensures that the decorated
    function always receives text string even
    if the decorator receives binary text.
    Once the wrapped function computes the result,
    decorator returns the same data-type as it
    has received.
    """

    def get_wrapped_object(self):
        def wrapped(string):
            is_binary = False
            if isinstance(string, six.binary_type):
                is_binary = True
                string = string.decode('utf-8')

            string = self.to_wrap(string)

            if is_binary:
                string = string.encode('utf-8')

            return string

        return wrapped


text_onion = TextOnion.as_decorator()



from __future__ import print_function, unicode_literals

import pytest
from django import forms

from django_auxilium.forms.color import ColorCharField


class TestColorCharField(object):
    """
    Test the color selector validation
    """

    def test_empty(self):
        field = ColorCharField(required=False)
        data = ['', None]
        for d in data:
            assert field.clean(d) == ''

    def test_type(self):
        valid_types = ['hex']
        for v in valid_types:
            field = ColorCharField(required=False, type=v)
            assert isinstance(field, ColorCharField)

        invalid_types = ['a', 'b']
        for i in invalid_types:
            with pytest.raises(ValueError):
                ColorCharField(required=False, type=i)

    def test_hex(self):
        field = ColorCharField()
        valid_values = ['#aaaaaa', 'aaaaaa']

        for v in valid_values:
            assert field.clean(v) == (v[1:].upper() if v[0] == '#' else v.upper())

        invalid_values = ['aa', '00ff0z']
        for i in invalid_values:
            with pytest.raises(forms.ValidationError):
                field.clean(i)

    def test_hex_hash_required(self):
        field = ColorCharField(required=False, hash_required=True)
        valid_values = ['#aaaaaa', '#bbbbbb']
        for v in valid_values:
            assert field.clean(v) == (v[1:].upper() if v[0] == '#' else v.upper())

        invalid_values = ['aaaaaa', 'bbbbbb']
        for i in invalid_values:
            with pytest.raises(forms.ValidationError):
                field.clean(i)

        field = ColorCharField(required=False, hash_required=False)
        data = valid_values + invalid_values
        for v in data:
            assert field.clean(v) == (v[1:].upper() if v[0] == '#' else v.upper())

    def test_hex_hash_output(self):
        field = ColorCharField(required=False, hash_output=True)
        data = ['#aaaaaa', '#bbbbbb', 'aaaaaa', 'bbbbbb']
        for v in data:
            assert field.clean(v) == (v.upper() if v[0] == '#' else '#' + v.upper())

        field = ColorCharField(required=False, hash_output=False)
        for v in data:
            assert field.clean(v) == (v[1:].upper() if v[0] == '#' else v.upper())



from __future__ import print_function, unicode_literals
from functools import WRAPPER_ASSIGNMENTS

import mock
from django.test import TestCase

from django_auxilium.utils.functools.decorators import (
    Decorator,
    HybridDecorator,
)


class TestDecorator(object):
    def test_get_wrapped_object(self):
        """
        Make sure in ``Decorator``, ``get_wrapped_object`` returns the input.
        """
        d = Decorator()
        d.to_wrap = mock.sentinel.object

        assert d.get_wrapped_object() is mock.sentinel.object

    def test_as_decorator(self):
        """
        Make sure ``to_decorator`` normalizes the wrapper
        """
        d = Decorator.as_decorator()

        def f():
            pass

        assert f is d(f)
        assert f is d()(f)

        class F(object):
            pass

        assert F is d(F)
        assert F is d()(F)

    def test_wraps(self):
        """
        Make sure the decorator correctly wraps the input function by
        preserving private attributes like ``__doc__``.
        """

        class D(Decorator):
            def get_wrapped_object(self):
                return lambda a: a

        d = D.as_decorator()

        def f():
            """
            Foo here
            """
            pass

        g = d(f)
        h = d()(f)

        for i in WRAPPER_ASSIGNMENTS:
            assert f is not g
            assert f is not h
            assert getattr(f, i) == getattr(g, i)
            assert getattr(f, i) == getattr(h, i)

    def test_call(self):
        """
        Make sure ``__call__`` returns the output of the wrapped function
        """
        d = Decorator()

        def f():
            return 'foo'

        g = d(f)

        assert f() is g()
        assert f is d.to_wrap


class TestHybridDecorator(TestCase):
    def setup_method(self, method):
        self.decorator = HybridDecorator()

    def test_pre_wrap_function(self):
        def foo():
            pass

        self.decorator(foo)

        assert not self.decorator.in_class

    def test_pre_wrap_function_with_parameters(self):
        def foo(bar):
            pass

        self.decorator(foo)

        assert not self.decorator.in_class

    def test_pre_wrap_method(self):
        def foo(self):
            pass

        self.decorator(foo)

        assert self.decorator.in_class

    def test_pre_wrap_provided(self):
        def foo(self):
            pass

        self.decorator.is_method = True
        self.decorator(foo)

        assert self.decorator.in_class



from __future__ import print_function

import six
from django.utils.functional import Promise

from django_auxilium.utils.functools import Decorator


SPECIAL_METHODS = [
    '__abs__',
    '__add__',
    '__and__',
    '__call__',
    '__cmp__',
    '__coerce__',
    '__contains__',
    '__delitem__',
    '__delslice__',
    '__div__',
    '__divmod__',
    '__eq__',
    '__float__',
    '__floordiv__',
    '__ge__',
    '__getitem__',
    '__getslice__',
    '__gt__',
    '__hash__',
    '__hex__',
    '__iadd__',
    '__iand__',
    '__idiv__',
    '__idivmod__',
    '__ifloordiv__',
    '__ilshift__',
    '__imod__',
    '__imul__',
    '__int__',
    '__invert__',
    '__ior__',
    '__ipow__',
    '__irshift__',
    '__isub__',
    '__iter__',
    '__itruediv__',
    '__ixor__',
    '__le__',
    '__len__',
    '__long__',
    '__lshift__',
    '__lt__',
    '__mod__',
    '__mul__',
    '__ne__',
    '__neg__',
    '__oct__',
    '__or__',
    '__pos__',
    '__pow__',
    '__radd__',
    '__rand__',
    '__rdiv__',
    '__rdivmod__',
    '__reduce__',
    '__reduce_ex__',
    '__repr__',
    '__reversed__',
    '__rfloorfiv__',
    '__rlshift__',
    '__rmod__',
    '__rmul__',
    '__ror__',
    '__rpow__',
    '__rrshift__',
    '__rshift__',
    '__rsub__',
    '__rtruediv__',
    '__rxor__',
    '__setitem__',
    '__setslice__',
    '__str__',
    '__sub__',
    '__truediv__',
    '__unicode__',
    '__xor__',
    'next',
]
"""
The list of special methods is compiled from
`<http://code.activestate.com/recipes/496741-object-proxying/>`_
"""


class LazyWrapperMeta(type(Promise)):
    """
    Metaclass for ``LazyWrapper``

    The reason we need metaclass is because we need to customize
    some magic methods for ``LazyWrapper`` for particular
    types while creating class which is what metaclasses are for!

    This metaclass expects all subclasses of ``LazyWrapper``
    to supply ``possible_types`` attribute which this
    metaclass will use to customize the created class.
    """

    def __new__(mcs, name, bases, attrs):
        _super = super(LazyWrapperMeta, mcs).__new__

        try:
            LazyWrapper

        except NameError:
            # defining LazyWrapper itself
            return _super(mcs, name, bases, attrs)

        else:
            possible_types = attrs.pop('possible_types')

            name += str('({})'.format(i.__name__ for i in possible_types))
            attrs.update(mcs._make_proxy_methods(possible_types))

            return _super(mcs, name, bases, attrs)

    @classmethod
    def _make_proxy_methods(mcs, possible_types):
        """
        Helper method for creating magic methods for proxy class
        for specific ``possible_types``.

        This is necessary because all magic methods of specified
        ``possible_types`` need to be replaced and since different data-types
        in Python implement different magic methods, we need to construct
        a dedicated class for each of the types and cannot simply
        reuse ``LazyWrapper`` itself.

        See Also
        --------
        _make_proxy_method
            That method explains in more detail why private magic
            methods need to be replaced.
        """
        methods = {}

        for t in possible_types:
            for name in SPECIAL_METHODS:
                if hasattr(t, name) and name not in methods:
                    methods[name] = mcs._make_proxy_method(name)

        return methods

    @staticmethod
    def _make_proxy_method(name):
        """
        This static method helps to create magic proxy methods
        which need to be overwritten on the proxy lazy class object.

        This is necessary because not all Python operations
        call ``__getattribute__``. For example doing ``print(obj)``
        never calls ``obj.__getattribute__('__str__')``
        but directly gets ``__str__`` which it then calls.
        Therefore simply overwriting ``__getattribute__``
        cant make objects lazy for all of the Python built-in
        operations and so the only solution is to explicitly
        replace such magic methods with versions which will
        evaluate lazy object first.
        """
        def method(self, *args, **kwargs):
            self._lazy_compute()
            return getattr(self._lazy_value, name)(*args, **kwargs)

        method.__name__ = str(name)
        return method


class LazyWrapper(six.with_metaclass(LazyWrapperMeta, Promise)):
    """
    Wrapper class for lazy objects as returned by the :py:class:`LazyDecorator`

    The job of this wrapper is to **not** execute the inner
    function to get the result until absolutely necessary.
    For example simply storing a variable will not trigger
    inner function to be executed until something is done
    with the variable such as printing it, etc.

    Parameters
    ----------
    f : def
        Function which this lazy object wraps
    args : tuple
        Tuple of arguments which were passed to the
        function for execution
    kwargs : dict
        Dict of keyword arguments which were passed to the
        function for execution
    """

    def __init__(self, f, args, kwargs):
        self._lazy_func = f
        self._lazy_args = args
        self._lazy_kwargs = kwargs
        self._lazy_value = None

    def _lazy_compute(self):
        """
        Compute the wrapped function and replace this method
        with noop so that future executions are fast
        """
        self._lazy_value = self._lazy_func(*self._lazy_args, **self._lazy_kwargs)
        self._lazy_computed = True
        self._lazy_compute = lambda: True

    def __getattribute__(self, name):
        """
        This method returns only lazy attributes from ``self``
        and for everything else returns attribute from the
        computed lazy object.
        """
        def get(x):
            return object.__getattribute__(self, x)

        # retrieve lazy attributes as normal
        if name.startswith('_lazy_'):
            return get(name)

        # get the attribute from the computed value
        get('_lazy_compute')()
        return get('_lazy_value').__getattribute__(name)

    def __setattr__(self, name, val):
        """
        This method allows to set lazy attributes on ``self``
        and for everything else sets them on computed
        lazy value.
        """
        # set lazy attributes normally
        if name.startswith('_lazy_'):
            object.__setattr__(self, name, val)

        else:
            self._lazy_compute()
            setattr(self._lazy_value, name, val)


class LazyDecorator(Decorator):
    """
    Lazy evaluation decorator.

    Parameters
    ----------
    types : tuple, list
        Iterable of possible output data-types of the output of the input function

    Examples
    --------

    ::

        >>> lazy = LazyDecorator

        >>> @lazy([six.string_types, int])
        ... def a(bar):
        ...   print('invoking a with {0}'.format(bar))
        ...   if isinstance(bar, int):
        ...       return bar + 5
        ...   else:
        ...       return bar + 'foo'

        >>> l = a('bar')
        >>> isinstance(l, six.string_types)
        invoking a with bar
        True
        >>> print(l)
        barfoo

        >>> l = a(5)
        >>> isinstance(l, int)
        invoking a with 5
        True
        >>> print(l)
        10
    """
    def __init__(self, types):
        self.types = self.get_possible_types(types)
        self.lazy_wrapper_class = self.get_lazy_wrapper_class()

    def get_possible_types(self, possible_types):
        """
        Get possible types as a list

        Parameters
        ----------
        possible_types : list, any
            Either a list of possible types or a single possible type

        Returns
        -------
        list
            List of possible types. If the input ``possible_types``
            was given as a single possible types, a list
            with a single item is returned.
        """
        return possible_types if isinstance(possible_types, (list, tuple)) else [possible_types]

    def get_lazy_wrapper_class(self):
        """
        Construct lazy wrapper class for the given possible types

        Returns
        -------
        type
            Lazy wrapper class specifically made for the
            possible types
        """
        return type(
            str('LazyWrapper'),
            (LazyWrapper,),
            {'possible_types': self.types}
        )

    def get_wrapped_object(self):
        """
        Get the wrapped callable which instead of computing
        values right away returns a lazy wrapper.
        """
        def wrapper(*args, **kwargs):
            return self.lazy_wrapper_class(self.to_wrap, args, kwargs)

        return wrapper


lazy = LazyDecorator


def flazy(f, possible_types):
    """
    Wrapper function for the :py:class:`LazyDecorator`

    It has the same API as Django's ``lazy`` function so the same code can be reused
    with this decorator if necessary.

    Examples
    --------

    ::

        >>> def f():
        ...     print('inside f')
        ...     return 5
        >>> g = flazy(f, int)
        >>> h = g()
        >>> print(str(h))
        inside f
        5
    """
    return LazyDecorator(possible_types)(f)



from __future__ import absolute_import, print_function, unicode_literals
import os
import uuid

import mock

from django_auxilium.models.fields.files import (
    OriginalFilenameRandomFileField,
    RandomFileField,
    original_random_filename_upload_to,
    random_filename_upload_to,
)


@mock.patch.object(uuid, 'uuid4')
@mock.patch.object(os.path, 'join')
def test_random_filename_upload_to(mock_join, mock_uuid4):
    mock_uuid4.return_value.hex = 'hex'

    upload_to = random_filename_upload_to('foo')
    result = upload_to(None, 'foo.jpg')

    assert result == mock_join.return_value
    mock_join.assert_called_once_with('foo', 'hex.jpg')


@mock.patch.object(uuid, 'uuid4')
@mock.patch.object(os.path, 'join')
def test_original_random_filename_upload_to(mock_join, mock_uuid4):
    mock_uuid4.return_value.hex = 'hex'
    instance = mock.Mock()

    upload_to = original_random_filename_upload_to('foo', 'filename')
    result = upload_to(instance, 'foo.jpg')

    assert result == mock_join.return_value
    assert instance.filename == 'foo.jpg'
    mock_join.assert_called_once_with('foo', 'hex.jpg')


class TestRandomFileField(object):
    def setup_method(self, method):
        self.field = RandomFileField(upload_to='foo')

    def test_init(self):
        assert self.field.upload_to_path == 'foo'
        assert callable(self.field.upload_to)


class TestOriginalFilenameRandomFileField(object):
    def setup_method(self, method):
        self.field = OriginalFilenameRandomFileField(upload_to='foo', filename_field='foo_filename')

    def test_init(self):
        assert self.field.upload_to_path == 'foo'
        assert self.field.filename_field == 'foo_filename'
        assert callable(self.field.upload_to)

    def test_deconstruct(self):
        actual = self.field.deconstruct()

        assert 'upload_to' in actual[3]
        assert actual[3]['upload_to'] == 'foo'
        assert 'filename_field' in actual[3]
        assert actual[3]['filename_field'] == 'foo_filename'



"""
Collection of simple utilities which modify some behaviour of functions
"""

from __future__ import print_function, unicode_literals
import inspect
import logging
from functools import partial, wraps


log = logging.getLogger(__name__)


class Decorator(object):
    """
    Base class for various decorators.

    Its aim is to do the same for decorators as Django class-based-views
    did to function-based-views.

    The most common pattern for creating decorators in Python can be summarized in the
    following snippet::

        from functools import wraps
        def decorator(f):
            @wraps(f)
            def wrapper(*args, **args):
                # some logic here
                return f(*args, **kwargs)
            return wrapper

    The problem with the above approach is when decorators need to employ more difficult
    logic. That sometimes can cause decorator functions to become big hence making it
    difficult to maintain and test. This class is meant to simplify this task. It provides
    a way to make decorators by using classes. Its aim is to do the same for decorators
    as Django class-based-views did to function-based-views. This means that decorator's various
    functionality can be split among class methods which can make the decorator's logic
    more transparent and more test-friendly. It also automatically wraps the decorated
    object using ``functools.wraps`` saving some coding lines. Finally, this class allows
    to create decorators where initializing the decorator is optional. This perhaps can
    better be illustrated in a snippet::

        @decorator             # not initialized
        def happy(): pass

        @decorator(foo='bar')  # initialized
        def bunnies(): pass

    Using this class is pretty simple. The most useful method you
    should know is :py:meth:`.get_wrapped_object`. Inside the method
    you can refer to :py:attr:`.to_wrap` which is the object the
    method should wrap. The method itself should return the wrapped
    version of the :py:attr:`.to_wrap` object. If your decorator
    needs to accept additional parameters, you can overwrite
    decorators ``__init__``. Finally, to use the decorator, don't
    forget to use the :py:meth:`as_decorator` class method.
    You can refer to **Examples** to see a more full
    example of how to use this class.

    Examples
    --------

    ::

        >>> class D(Decorator):
        ...     def __init__(self, happy=False):
        ...         self.happy = happy
        ...     def get_wrapped_object(self):
        ...         def wrapped(*args, **kwargs):
        ...             print('called wrapped with', args, kwargs)
        ...             if self.happy:
        ...                 print('Bunnies are very happy today!')
        ...             return self.to_wrap(*args, **kwargs)
        ...         return wrapped
        ...
        >>> decorator = D.as_decorator()

        >>> # not initialized (default values will be used if defined)
        >>> @decorator
        ... def sum(a, b):
        ...    return a + b

        >>> sum(5, 6)
        called wrapped with (5, 6) {}
        11

        >>> # initialized with no parameters (default values will be used if defined)
        >>> @decorator()
        ... def sum(a, b):
        ...    return a + b

        >>> sum(7, 8)
        called wrapped with (7, 8) {}
        15

        >>> # initialized with keyword parameters
        >>> @decorator(happy=True)
        ... def sum(a, b):
        ...    "Sum function"
        ...    return a + b

        >>> sum(9, 10)
        called wrapped with (9, 10) {}
        Bunnies are very happy today!
        19

        >>> sum.__doc__ == 'Sum function'
        True

    Attributes
    ----------
    to_wrap : object
        The object to be decorated/wrapped. This attributes becomes available when the
        decorator is called on the object.
    """

    def __call__(self, f):
        self.to_wrap = f

        self.pre_wrap()
        if inspect.isclass(f):
            return self.get_wrapped_object()
        else:
            return wraps(f)(self.get_wrapped_object())

    def pre_wrap(self):
        """
        Hook for executing things before wrapping objects
        """

    def get_wrapped_object(self):
        """
        Returns the wrapped version of the object to be decorated/wrapped.

        This is the meat of class because this method is the one which creates the
        decorator. Inside the method, you can refer to the object to be decorated via
        ``self.to_wrap``.

        .. note::
            This class automatically uses the ``functools.wraps`` to preserve the
            ``to_wrap``'s object useful attributes such as ``__doc__`` hence there
            is no need to do that manually. You can just return a wrapped object
            and the class will take care of the rest.

        Returns
        -------
        f : object
            Decorated/wrapped function
        """
        return self.to_wrap

    @classmethod
    def as_decorator(cls, *a, **kw):
        """
        Return the actual decorator.

        This method is necessary because the decorator is a class. Consider the
        following::

            >>> class ClassDecorator(object):
            ...     def __init__(self, obj):
            ...         self.to_wrap = obj

            >>> @ClassDecorator
            ... def foo(): pass
            >>> isinstance(foo, ClassDecorator)
            True

        In the above, since ``foo`` will be passed to the decorator's class ``__init__``,
        the returned object will be an instance of the decorator's class instead of a
        wrapper function. To avoid that, this method constructs a wrapper function
        which guarantees that the output of the decorator will be the wrapped object::

            >>> class D(Decorator):
            ...     pass
            >>> d = D.as_decorator()
            >>> @d
            ... def foo(): pass
            >>> isinstance(foo, D)
            False
        """
        klass = partial(cls, *a, **kw)

        @wraps(cls)
        def wrapper(*args, **kwargs):
            if args and (callable(args[0]) or inspect.isclass(args[0])):
                return klass()(args[0])
            else:
                return klass(*args, **kwargs)

        return wrapper


class HybridDecorator(Decorator):
    """
    Class for implementing decorators which can decorate both regular functions as well
    as class methods.

    This decorator automatically determines if the object to be decorated is a stand-alone
    function or a class method by adding an :py:attr:`in_class` attribute.

    Parameters
    ----------
    is_method : bool, optional
        Explicitly specify whether the decorator is being used on class
        method or a standalone function.
        When not specified, :py:meth:`.pre_wrap` is used to automatically
        determine that. Please look at its documentation for the
        explanation of its limitations.

    Attributes
    ----------
    in_class : bool
        ``True`` if the object to be decorated is a class method, or ``False`` if it is a
        standalone function
    """

    def __init__(self, is_method=None):
        self.is_method = None

    def pre_wrap(self):
        """
        Method which determines whether the ``to_wrap`` is a class
        method or a standalone function

        .. warning::
            This method uses pretty primitive technique to determine
            whether the wrapped callable is a class method or a function
            and so it might not work in all cases.
            It checks the first parameter name of the callable
            and if it is either ``'self'`` or ``'cls'`` it is
            most likely a method.

            If you need more precise behavior you are encouraged to
            use ``is_method`` decorator parameter.
        """
        if self.is_method is None:
            try:
                arg = inspect.getargspec(self.to_wrap).args[0]
            except IndexError:
                self.in_class = False
            else:
                self.in_class = arg in ['self', 'cls']
        else:
            self.in_class = self.is_method



from __future__ import print_function, unicode_literals

import mock
import pytest
from dirtyfields import DirtyFieldsMixin
from django.core.files.base import ContentFile
from django.db import models
from django.test import TestCase

from django_auxilium.models import (
    AutoSignals,
    FieldSpec,
    FileFieldAutoChangeDelete,
    FileFieldAutoDelete,
)


class TestFileFieldAutoDelete(object):
    def setup_method(self, method):
        self.decorator = FileFieldAutoDelete()

    def get_model(self):
        class Model(models.Model):
            file_field = models.FileField(upload_to='foo')
            foo_field = models.CharField(max_length=32)

            class Meta(object):
                app_label = 'foo'

        return Model

    def get_empty_model(self):
        class ModelEmpty(models.Model):
            class Meta(object):
                app_label = 'foo'

        return ModelEmpty

    def test_init(self):
        decorator = FileFieldAutoDelete()

        assert decorator.field_names == '*'
        assert decorator.signal is models.signals.post_delete
        assert decorator.signal_name_pattern == 'post_delete_{model.__name__}_delete_{field}'

    @mock.patch.object(FileFieldAutoDelete, 'connect_signal_function')
    @mock.patch.object(FileFieldAutoDelete, 'validate_fields')
    @mock.patch.object(FileFieldAutoDelete, 'get_field_names')
    @mock.patch.object(FileFieldAutoDelete, 'get_fields')
    @mock.patch.object(FileFieldAutoDelete, 'validate_model')
    def test_get_wrapped_object(self, mock_validate_model,
                                mock_get_fields, mock_get_field_names,
                                mock_validate_fields,
                                mock_connect_signal_function):
        self.decorator.to_wrap = mock.sentinel.model

        # sanity assert
        assert self.decorator.field_names == '*'
        assert self.decorator.get_wrapped_object() is mock.sentinel.model
        assert self.decorator.field_names == mock_get_field_names.return_value
        assert self.decorator.fields == mock_get_fields.return_value

        mock_validate_model.assert_called_once_with()
        mock_get_field_names.assert_called_once_with()
        mock_get_fields.assert_called_once_with()
        mock_validate_fields.assert_called_once_with()
        mock_connect_signal_function.assert_called_once_with()

    def test_validate_model_not_class(self):
        self.decorator.to_wrap = None

        with pytest.raises(TypeError):
            self.decorator.validate_model()

    def test_validate_model_not_model(self):
        self.decorator.to_wrap = int

        with pytest.raises(TypeError):
            self.decorator.validate_model()

    def test_validate_model(self):
        self.decorator.to_wrap = self.get_empty_model()

        assert self.decorator.validate_model() is None

    def test_get_field_names_single_field(self):
        self.decorator.field_names = 'foo'

        assert self.decorator.get_field_names() == ['foo']

    def test_get_field_names_single_multiple_fields(self):
        self.decorator.field_names = ['foo', 'bar']

        assert self.decorator.get_field_names() == ['foo', 'bar']

    def test_get_field_names_single_wildcard(self):
        self.decorator.field_names = '*'
        self.decorator.to_wrap = self.get_model()

        assert self.decorator.get_field_names() == ['file_field']

    @mock.patch.object(FileFieldAutoDelete, 'validate_field')
    def test_validate_fields(self, mock_validate_field):
        self.decorator.field_names = ['foo']

        assert self.decorator.validate_fields() is None

        mock_validate_field.assert_called_once_with('foo')

    def test_validate_field_not_present(self):
        self.decorator.to_wrap = self.get_model()

        with pytest.raises(AttributeError):
            self.decorator.validate_field('foo')

    def test_validate_field_not_filefield(self):
        self.decorator.to_wrap = self.get_model()

        with pytest.raises(TypeError):
            self.decorator.validate_field('foo_field')

    def test_validate_field(self):
        self.decorator.to_wrap = self.get_model()

        assert self.decorator.validate_field('file_field') is None

    def test_get_signal_name_single(self):
        self.decorator.to_wrap = self.get_model()
        self.decorator.fields = [FieldSpec('foo', None)]

        name = self.decorator.get_signal_name()
        assert name == 'post_delete_Model_delete_foo'

    def test_get_signal_name_multiple(self):
        self.decorator.to_wrap = self.get_model()
        self.decorator.fields = [FieldSpec('foo', None), FieldSpec('bar', None)]

        name = self.decorator.get_signal_name()
        assert name == 'post_delete_Model_delete_foo_bar'

    def test_get_signal_function_no_file(self):
        model = self.get_model()
        self.decorator.to_wrap = model
        self.decorator.fields = [FieldSpec('file_field', None)]

        signal = self.decorator.get_signal_function()

        assert callable(signal)

        f = mock.MagicMock(spec=ContentFile)
        instance = model(file_field=f)

        assert signal(model, instance) is None
        assert not f.mock_calls

    def test_get_signal_function_with_file(self):
        model = self.get_model()
        self.decorator.to_wrap = model
        self.decorator.fields = [FieldSpec('file_field', None)]

        signal = self.decorator.get_signal_function()

        assert callable(signal)

        f = mock.MagicMock()
        instance = model(file_field=f)

        assert signal(model, instance) is None
        f.delete.assert_called_once_with(save=False)

    @mock.patch.object(FileFieldAutoDelete, 'get_signal_function')
    def test_connect_signal_function(self, mock_get_signal_function):
        self.decorator.signal = signal = mock.MagicMock()
        self.decorator.fields = [FieldSpec('foo', None)]
        self.decorator.to_wrap = model = self.get_model()

        assert self.decorator.connect_signal_function() is None
        signal.connect.assert_called_once_with(
            mock_get_signal_function.return_value,
            sender=model,
            weak=False,
            dispatch_uid='post_delete_Model_delete_foo'
        )


class TestFileFieldAutoChangeDelete(TestCase):
    def setup_method(self, method):
        self.decorator = FileFieldAutoChangeDelete()

    def get_model(self):
        class Model(DirtyFieldsMixin, models.Model):
            file_field = models.FileField(upload_to='foo')
            foo_field = models.CharField(max_length=32)

            class Meta(object):
                app_label = 'foo'

        return Model

    def get_empty_model(self):
        class ModelEmpty(DirtyFieldsMixin, models.Model):
            class Meta(object):
                app_label = 'foo'

        return ModelEmpty

    def test_init(self):
        decorator = FileFieldAutoChangeDelete()

        assert decorator.field_names == '*'
        assert decorator.signal is models.signals.post_save
        assert decorator.signal_name_pattern == 'post_save_{model.__name__}_changedelete_{field}'

    def test_validate_model_no_is_dirty_implementation(self):
        self.decorator.to_wrap = self.get_empty_model()
        self.decorator.to_wrap.is_dirty = None
        self.decorator.to_wrap.get_dirty_fields = None

        with pytest.raises(ValueError):
            self.decorator.validate_model()

    def test_validate_model(self):
        self.decorator.to_wrap = self.get_empty_model()

        assert self.decorator.validate_model() is None

    @mock.patch.object(DirtyFieldsMixin, 'get_dirty_fields')
    @mock.patch.object(DirtyFieldsMixin, 'is_dirty')
    def test_get_signal_function_not_dirty(
            self, mock_is_dirty, mock_get_dirty_fields):
        mock_is_dirty.return_value = False
        model = self.get_model()
        self.decorator.to_wrap = model
        self.decorator.fields = ['file_field']

        signal = self.decorator.get_signal_function()

        assert callable(signal)

        instance = model()

        assert signal(model, instance) is None
        mock_is_dirty.assert_called_once_with()
        assert not mock_get_dirty_fields.called

    @mock.patch.object(DirtyFieldsMixin, 'get_dirty_fields')
    @mock.patch.object(DirtyFieldsMixin, 'is_dirty')
    def test_get_signal_function_dirty_but_no_file(
            self, mock_is_dirty, mock_get_dirty_fields):
        mock_is_dirty.return_value = True
        mock_get_dirty_fields.return_value = {}
        model = self.get_model()
        self.decorator.to_wrap = model
        self.decorator.fields = [FieldSpec('file_field', None)]

        signal = self.decorator.get_signal_function()

        assert callable(signal)

        instance = model()

        assert signal(model, instance) is None
        mock_is_dirty.assert_called_once_with()
        mock_get_dirty_fields.assert_called_once_with()

    @mock.patch.object(DirtyFieldsMixin, 'get_dirty_fields')
    @mock.patch.object(DirtyFieldsMixin, 'is_dirty')
    def test_get_signal_function(
            self, mock_is_dirty, mock_get_dirty_fields):
        f = mock.MagicMock()
        field = mock.MagicMock()
        mock_is_dirty.return_value = True
        mock_get_dirty_fields.return_value = {'file_field': f}
        model = self.get_model()
        self.decorator.to_wrap = model
        self.decorator.fields = [FieldSpec('file_field', field)]

        signal = self.decorator.get_signal_function()

        assert callable(signal)

        instance = model(file_field='foo')

        assert signal(model, instance) is None
        mock_is_dirty.assert_called_once_with()
        mock_get_dirty_fields.assert_called_once_with()
        f.delete.assert_called_once_with(save=False)
        assert f.instance == instance
        assert f.field == field
        assert f.storage == field.storage
        assert instance.file_field == 'foo'


class TestAutoSignals(object):
    def setup_method(self, method):
        self.decorator = AutoSignals()

    def test_init(self):
        supported_signals = {
            'class_prepared',
            'm2m_changed',
            'post_delete',
            'post_init',
            'post_save',
            'pre_delete',
            'pre_init',
            'pre_save',
        }
        assert self.decorator.getter == 'get_signals'
        assert isinstance(self.decorator.signal_pool, dict)
        assert set(self.decorator.signal_pool) & supported_signals == supported_signals

    def test_validate_model_not_class(self):
        self.decorator.to_wrap = None
        with pytest.raises(TypeError):
            self.decorator.validate_model()

    def test_validate_model_not_django_model(self):
        self.decorator.to_wrap = int
        with pytest.raises(TypeError):
            self.decorator.validate_model()

    def test_validate_model_not_implementing_getter(self):
        class Foo(models.Model):
            class Meta(object):
                app_label = 'foo'

        self.decorator.to_wrap = Foo
        with pytest.raises(AttributeError):
            self.decorator.validate_model()

    def test_validate_model_not_implementing_getter_as_callable(self):
        class Foo(models.Model):
            get_signals = 'foo'

            class Meta(object):
                app_label = 'foo'

        self.decorator.to_wrap = Foo
        with pytest.raises(TypeError):
            self.decorator.validate_model()

    def test_validate_model_valid(self):
        class Foo(models.Model):
            class Meta(object):
                app_label = 'foo'

            @classmethod
            def get_signals(self):
                return []

        self.decorator.to_wrap = Foo
        assert self.decorator.validate_model() is None

    @mock.patch.object(AutoSignals, 'validate_model')
    @mock.patch.object(AutoSignals, 'connect_signals')
    def test_get_wrapper_object(self, mock_connect_signals, mock_validate_model):
        self.decorator.to_wrap = mock.sentinel.to_wrap

        assert self.decorator.get_wrapped_object() is mock.sentinel.to_wrap
        mock_connect_signals.assert_called_once_with()
        mock_validate_model.assert_called_once_with()

    @mock.patch.object(AutoSignals, 'connect_signal')
    def test_connect_signals_list(self, mock_connect_signal):
        def pre_save(sender, instance, *args, **kwwargs):
            pass

        class Foo(models.Model):
            class Meta(object):
                app_label = 'foo'

            @classmethod
            def get_signals(self):
                return [pre_save]

        self.decorator.to_wrap = Foo

        self.decorator.connect_signals()
        mock_connect_signal.assert_called_once_with(pre_save)

    @mock.patch.object(AutoSignals, 'connect_signal')
    def test_connect_signals_single(self, mock_connect_signal):
        def pre_save(sender, instance, *args, **kwwargs):
            pass

        class Foo(models.Model):
            class Meta(object):
                app_label = 'foo'

            @classmethod
            def get_signals(self):
                return pre_save

        self.decorator.to_wrap = Foo

        self.decorator.connect_signals()
        mock_connect_signal.assert_called_once_with(pre_save)

    def test_connect_signal_invalid_type(self):
        self.decorator.to_wrap = mock.sentinel.model

        with pytest.raises(TypeError):
            self.decorator.connect_signal(None)

    def test_connect_signal_no_receiver(self):
        self.decorator.to_wrap = mock.sentinel.model

        with pytest.raises(ValueError):
            self.decorator.connect_signal({})

    def test_connect_signal_cant_determine_signal(self):
        self.decorator.to_wrap = mock.sentinel.model

        def foo(*args, **kwargs):
            pass

        with pytest.raises(ValueError):
            self.decorator.connect_signal(foo)

    def test_connect_signal_invalid_signal(self):
        self.decorator.to_wrap = mock.sentinel.model

        def foo(*args, **kwargs):
            pass

        with pytest.raises(ValueError):
            self.decorator.connect_signal({
                'receiver': foo,
                'signal': 'foo',
            })

    def test_connect_signal_from_callable(self):
        mock_signal = mock.MagicMock()
        self.decorator.to_wrap = mock.sentinel.model
        self.decorator.signal_pool = {'pre_save': mock_signal}

        def foo_pre_save(*args, **kwargs):
            pass

        self.decorator.connect_signal(foo_pre_save)

        mock_signal.connect.assert_called_once_with(
            sender=mock.sentinel.model,
            receiver=foo_pre_save,
        )

    def test_connect_signal_from_dict(self):
        mock_signal = mock.MagicMock()
        self.decorator.to_wrap = mock.sentinel.model
        self.decorator.signal_pool = {'pre_save': mock_signal}

        def foo_pre_save(*args, **kwargs):
            pass

        self.decorator.connect_signal({
            'receiver': foo_pre_save,
            'weak': False,
        })

        mock_signal.connect.assert_called_once_with(
            sender=mock.sentinel.model,
            receiver=foo_pre_save,
            weak=False,
        )



from __future__ import absolute_import, print_function, unicode_literals

import six

from django_auxilium.utils.string import text_onion


def test_text_onion():
    @text_onion
    def foo(a):
        assert isinstance(a, six.text_type)
        return a

    assert foo('foo') == 'foo'
    assert foo(b'foo') == b'foo'



from __future__ import print_function, unicode_literals

from django_auxilium.utils.html import (
    TextExtractorHTMLParser,
    html_to_text,
    simple_minify,
)


MINIFY_INPUT = """
<html>
<head>
    <title>Minify Test</title>
    <script>
        (function() {
            console.log('hello world');
        })();
    </script>
</head>
<body>
    <div>Content Here</div>
    <textarea>
Input
Here
  123
    </textarea>
    <code>Code
    Here</code>
    <pre>
    <span>Hello</span>
    <b>World</b>
    {
        "json": "here"
    }
    </pre>
    <script>
        (function() {
            console.log('inside body script');
        })();
    </script>
</body>
</html>
"""

MINIFY_EXPECTED = """<html><head><title>Minify Test</title><script>
        (function() {
            console.log('hello world');
        })();
    </script></head><body><div>Content Here</div><textarea>
Input
Here
  123
    </textarea><code>Code
    Here</code><pre>
    <span>Hello</span>
    <b>World</b>
    {
        "json": "here"
    }
    </pre><script>
        (function() {
            console.log('inside body script');
        })();
    </script></body></html>"""


EXTRACT_INPUT = """
<html>
<head>
    <title>Minify Test</title>
</head>
<body>
    <div>Content Here</div>
    &gt;&#62;&#x3E;
</body>
</html>
"""

EXTRACT_EXPECTED = """Minify Test


    Content Here
    >>>"""


def test_simple_minify():
    assert simple_minify(MINIFY_INPUT) == MINIFY_EXPECTED


def test_html_to_text():
    assert html_to_text(EXTRACT_INPUT).strip() == EXTRACT_EXPECTED


def test_handle_charref():
    parser = TextExtractorHTMLParser()
    assert parser.handle_charref('62') == '>'


def test_handle_entityref():
    parser = TextExtractorHTMLParser()
    assert parser.handle_entityref('gt') == '>'



from __future__ import absolute_import, print_function, unicode_literals

import six

from django_auxilium.utils.functools.lazy import LazyWrapper, flazy, lazy


class Foo(six.text_type):
    pass


def test_lazy():
    @lazy(six.text_type)
    def foo():
        return Foo('foo')

    f = foo()

    assert issubclass(type(f), LazyWrapper)
    assert f == 'foo'
    assert f.startswith('f')

    f.foo = 'hello'
    assert f.foo == 'hello'


def test_flazy():
    def _foo():
        return Foo('foo')

    foo = flazy(_foo, six.text_type)

    f = foo()

    assert issubclass(type(f), LazyWrapper)
    assert f == 'foo'
    assert f.startswith('f')

    f.foo = 'hello'
    assert f.foo == 'hello'



from __future__ import print_function, unicode_literals

from django.core.mail.backends.filebased import EmailBackend

from .base import BaseCallbackEmailBackendMixin


class CallbackEmailBackend(BaseCallbackEmailBackendMixin, EmailBackend):
    """
    Custom file-based email backend which sends signals on progress events
    of sending messages.
    """

    def send_messages(self, email_messages):
        """
        Custom implementation of sending messages which also
        calls mail signals.

        Unfortunately vanilla Django implementation does not
        provide enough hooks for when messages are sent
        hence this method needs to duplicate the whole method.
        """
        if not email_messages:
            return
        msg_count = 0
        with self._lock:
            try:
                stream_created = self.open()
                for message in email_messages:
                    self.sending(message)
                    self.write_message(message)
                    self.sent(message)
                    msg_count += 1
                if stream_created:
                    self.close()
            except Exception:
                for message in email_messages:
                    self.failed(message, 'Could not write emails to file')
                if not self.fail_silently:
                    raise
        return msg_count



from __future__ import print_function, unicode_literals

import pytest

from django_auxilium.utils.range import AlphabeticNumbers, Range


class TestAlphabeticNumbers(object):
    def test_int_from_str(self):
        data = [
            ('A', 1),
            ('Z', 26),
            ('AA', 27),
            ('BA', 53),
            ('AAA', 703),
            ('XZZ', 16926),
            ('ZZZ', 18278),
        ]
        for d in data:
            assert AlphabeticNumbers.int_from_str(d[0]) == d[1]

    def test_str_from_int(self):
        data = [
            ('A', 1),
            ('Z', 26),
            ('AA', 27),
            ('BA', 53),
            ('AAA', 703),
            ('XZZ', 16926),
            ('ZZZ', 18278),
        ]
        for d in data:
            assert AlphabeticNumbers.str_from_int(d[1]) == d[0]

        for i in [0, -1]:
            with pytest.raises(ValueError):
                AlphabeticNumbers.str_from_int(i)


def test_range_columns():
    assert Range(0, 1, 2, 3).columns == 0
    assert Range(1, 1, 2, 3).columns == 2


def test_range_rows():
    assert Range(0, 0, 2, 3).rows == 0
    assert Range(0, 1, 2, 2).rows == 2



"""
Collection of Django custom model field which have something to do with files
"""

from __future__ import print_function, unicode_literals
import os
import uuid

from django.db import models


def random_filename_upload_to(path):
    """
    Get ``upload_to`` compliant method which will always return a random filename.

    This method can be used as Django's ``FileField`` ``upload_to`` parameter.
    It returns an ``upload_to`` callable which will make sure the filename will always
    be random. This method also accepts a ``path`` parameter which will define where
    relative to the media folder, the file will be stored.

    Parameters
    ----------
    path : str
        The path relative to ``media`` where the filename should be uploaded to

    Returns
    -------
    upload_to : function
        Django ``FileField``'s ``upload_to`` compliant function
    """

    def f(instance, filename):
        ext = filename.split('.')[-1]
        filename = '{0}.{1}'.format(uuid.uuid4().hex, ext)
        return os.path.join(path, filename)

    return f


def original_random_filename_upload_to(path, filename_field):
    """
    Get ``upload_to`` compliant method which will always return a random filename
    however will also store the original filename in the model.

    Parameters
    ----------
    path : str
        The path relative to ``media`` where the filaname should be uploaded to
    filename_field : str
        The name of the model attribute to which the original filename should be stored.

    Returns
    -------
    upload_to : function
        Django ``FileField``'s ``upload_to`` compliant function
    """
    g = random_filename_upload_to(path)

    def f(instance, filename):
        setattr(instance, filename_field, filename)
        return g(instance, filename)

    return f


class RandomFileFieldDeconstructMixin(object):
    """
    Mixin for random filename file mixins which implements
    Django's ``deconstruct`` to be Django-migrations compatible
    """

    def deconstruct(self):
        """
        Standard Django's ``deconstruct`` in order to enable migrations support
        """
        name, path, args, kwargs = super(RandomFileFieldDeconstructMixin, self).deconstruct()
        kwargs.update({
            'upload_to': self.upload_to_path,
        })
        if hasattr(self, 'filename_field') and self.filename_field != 'filename':
            kwargs.update({
                'filename_field': self.filename_field,
            })
        return name, path, args, kwargs


class RandomFileNameFileFieldMixin(RandomFileFieldDeconstructMixin):
    """
    Mixing for a custom ``FileField`` which generates random filename

    Parameters
    ----------
    upload_to : str
        String where uploaded files should be saved.
        Within that directory, random filename will always be selected.
    """

    def __init__(self, *args, **kwargs):
        assert kwargs.get('upload_to') is not None

        self.upload_to_path = kwargs['upload_to']
        kwargs['upload_to'] = random_filename_upload_to(kwargs['upload_to'])

        super(RandomFileNameFileFieldMixin, self).__init__(*args, **kwargs)


class RandomFileNameWithFilenameFileFieldMixin(RandomFileFieldDeconstructMixin):
    """
    Mixing for a custom ``FileField`` which generates random filename
    and also stores original filename

    Parameters
    ----------
    upload_to : str
        String where uploaded files should be saved.
        Within that directory, random filename will always be selected.
    filename_field : str
        Name of Django model field name where original filename
        should be stored
    """
    def __init__(self, *args, **kwargs):
        assert kwargs.get('upload_to') is not None

        self.filename_field = kwargs.pop('filename_field', 'filename')
        self.upload_to_path = kwargs['upload_to']
        kwargs['upload_to'] = original_random_filename_upload_to(
            kwargs['upload_to'],
            self.filename_field
        )

        super(RandomFileNameWithFilenameFileFieldMixin, self).__init__(*args, **kwargs)


RandomFileField = type(
    str('RandomFileField'),
    (RandomFileNameFileFieldMixin, models.FileField,),
    {}
)

OriginalFilenameRandomFileField = type(
    str('OriginalFilenameRandomFileField'),
    (RandomFileNameWithFilenameFileFieldMixin, models.FileField,),
    {}
)

RandomImageField = type(
    str('RandomImageField'),
    (RandomFileNameFileFieldMixin, models.ImageField,),
    {}
)

OriginalFilenameRandomImageField = type(
    str('OriginalFilenameRandomImageField'),
    (RandomFileNameWithFilenameFileFieldMixin, models.ImageField,),
    {}
)



from __future__ import absolute_import, print_function
import types
from functools import partial

import pytest

from django_auxilium.utils.functools.cache import (
    CacheDecorator,
    CacheDescriptor,
    Caching,
    MemoizeDecorator,
    MemoizeDescriptor,
    Memoizing,
    NotInCache,
)


class Bunch(object):
    pass


class TestCaching(object):
    def setup_method(self, method):
        self.object = Bunch()
        self.cache = Caching(self.object, 'cache')

    def test_get_not_present(self):
        with pytest.raises(NotInCache):
            self.cache.get()

    def test_get_present(self):
        self.object.cache = 'foo'

        assert self.cache.get() == 'foo'

    def test_delete_not_present(self):
        with pytest.raises(NotInCache):
            self.cache.delete()

    def test_delete_present(self):
        self.object.cache = 'foo'

        assert self.cache.delete() == 'foo'
        assert not hasattr(self.object, 'cache')

    def test_set(self):
        assert self.cache.set('foo') == 'foo'
        assert self.object.cache == 'foo'


class TestMemoizing(object):
    def setup_method(self, method):
        self.object = Bunch()
        self.cache = Memoizing(self.object, 'cache')
        self.key = "('foo',)[]"

    def test_get_key(self):
        assert self.cache._get_key('foo') == "('foo',)[]"
        assert self.cache._get_key('foo', 'bar') == "('foo', 'bar')[]"
        assert self.cache._get_key(foo='bar') == "()[('foo', 'bar')]"
        assert self.cache._get_key(foo='bar', hello='there') == "()[('foo', 'bar'), ('hello', 'there')]"
        assert self.cache._get_key('foo', foo='bar') == "('foo',)[('foo', 'bar')]"

    def test_get_not_present(self):
        with pytest.raises(NotInCache):
            self.cache.get('foo')

    def test_get_present(self):
        self.object.cache = {self.key: 'foo'}

        assert self.cache.get('foo') == 'foo'

    def test_delete_not_present(self):
        with pytest.raises(NotInCache):
            self.cache.delete()

    def test_delete_present(self):
        self.object.cache = {self.key: 'foo'}

        assert self.cache.delete('foo') == 'foo'
        assert self.object.cache == {}

    def test_set(self):
        assert self.cache.set('foo', 'foo') == 'foo'
        assert self.object.cache == {self.key: 'foo'}


class TestCacheDescriptor(object):
    def setup_method(self, method):
        def bar(self):
            """
            hello world
            """
            return 'bar'

        class Foo(object):
            foo = CacheDescriptor(bar)

        self.bar = bar
        self.klass = Foo
        self.instance = Foo()
        self.descriptor = Foo.__dict__['foo']

    def test_init(self):
        assert self.descriptor.method is self.bar
        assert self.descriptor.cache_class is Caching
        assert not self.descriptor.as_property
        assert self.descriptor.cache_attribute.startswith('bar_cache_')

    def test_get_cache(self):
        actual = self.descriptor.get_cache(self.instance)

        assert isinstance(actual, Caching)
        assert actual.parent is self.instance
        assert actual.attr == self.descriptor.cache_attribute

    def test_getter_cache_present(self):
        setattr(self.instance, self.descriptor.cache_attribute, 'haha')

        actual = self.descriptor.getter(self.instance)

        assert actual == 'haha'
        assert hasattr(self.instance, self.descriptor.cache_attribute)
        assert getattr(self.instance, self.descriptor.cache_attribute) == 'haha'

    def test_getter_cache_not_present(self):
        assert not hasattr(self.instance, self.descriptor.cache_attribute)

        actual = self.descriptor.getter(self.instance)

        assert actual == 'bar'
        assert hasattr(self.instance, self.descriptor.cache_attribute)
        assert getattr(self.instance, self.descriptor.cache_attribute) == 'bar'

    def test_pop_cache_present(self):
        setattr(self.instance, self.descriptor.cache_attribute, 'haha')

        actual = self.descriptor.pop(self.instance)

        assert actual == 'haha'
        assert not hasattr(self.instance, self.descriptor.cache_attribute)

    def test_pop_cache_not_present(self):
        assert not hasattr(self.instance, self.descriptor.cache_attribute)

        with pytest.raises(NotInCache):
            self.descriptor.pop(self.instance)

    def test_get_class(self):
        assert self.klass.foo is self.bar

    def test_get_instance(self):
        assert isinstance(self.instance.foo, partial)
        assert isinstance(self.instance.foo.func, types.MethodType)
        assert self.instance.foo.func.__self__ is self.instance

        assert isinstance(self.instance.foo.pop, partial)
        assert isinstance(self.instance.foo.pop.func, types.MethodType)
        assert self.instance.foo.pop.func.__self__ is self.instance

        assert self.instance.foo.__doc__ == self.bar.__doc__

        assert self.instance.foo() == 'bar'

    def test_delete(self):
        with pytest.raises(AttributeError):
            del self.instance.foo

    def test_get_class_property(self):
        self.descriptor.as_property = True

        assert self.klass.foo is self.descriptor

    def test_get_instance_property(self):
        self.descriptor.as_property = True

        assert self.instance.foo == 'bar'

    def test_set_read_only_method(self):
        self.descriptor.as_property = False

        with pytest.raises(AttributeError):
            self.instance.foo = 'foo'

    def test_set_writable_property(self):
        self.descriptor.as_property = True

        self.instance.foo = 'foo'

        assert self.instance.foo == 'foo'

    def test_delete_property_not_present(self):
        self.descriptor.as_property = True

        with pytest.raises(NotInCache):
            del self.instance.foo

    def test_delete_property_present(self):
        self.descriptor.as_property = True
        setattr(self.instance, self.descriptor.cache_attribute, 'haha')

        del self.instance.foo

        assert not hasattr(self.instance, self.descriptor.cache_attribute)


class TestMemoizeDescriptor(object):
    def setup_method(self, method):
        def bar(self, a):
            """
            hello world
            """
            return 'bar'

        class Foo(object):
            foo = MemoizeDescriptor(bar)

        self.bar = bar
        self.klass = Foo
        self.instance = Foo()
        self.descriptor = Foo.__dict__['foo']

    def test_init(self):
        assert self.descriptor.method is self.bar
        assert self.descriptor.cache_class is Memoizing
        assert self.descriptor.cache_attribute.startswith('bar_memoize_')


class TestCacheDecorator(object):
    def test_function(self):
        self.counter = 0

        @CacheDecorator()
        def foo():
            self.counter += 1
            return self.counter

        assert isinstance(foo.decorator, CacheDecorator)
        assert isinstance(foo.decorator.cache, Caching)
        assert foo.decorator.cache.parent is foo.decorator
        assert foo.decorator.cache.attr == 'cached_value'

        assert hasattr(foo, 'pop')
        assert foo.pop.__self__ is foo.decorator

        assert foo() == 1
        assert foo() == 1
        assert foo.pop() == 1
        assert foo() == 2

    def test_method(self):
        class Foo(object):
            def __init__(self):
                self.counter = 0

            @CacheDecorator()
            def foo(self):
                self.counter += 1
                return self.counter

        descriptor = Foo.__dict__['foo']

        assert isinstance(descriptor, CacheDescriptor)
        assert not descriptor.as_property

    def test_property(self):
        class Foo(object):
            def __init__(self):
                self.counter = 0

            @CacheDecorator(as_property=True)
            def foo(self):
                self.counter += 1
                return self.counter

        descriptor = Foo.__dict__['foo']

        assert isinstance(descriptor, CacheDescriptor)
        assert descriptor.as_property


class TestMemoizeDecorator(object):
    def test_function(self):
        self.counter = 0

        @MemoizeDecorator()
        def foo(a):
            self.counter += 1
            return self.counter

        assert isinstance(foo.decorator, MemoizeDecorator)
        assert isinstance(foo.decorator.cache, Memoizing)
        assert foo.decorator.cache.parent is foo.decorator
        assert foo.decorator.cache.attr == 'cached_value'

        assert hasattr(foo, 'pop')
        assert foo.pop.__self__ is foo.decorator

        assert foo('a') == 1
        assert foo('b') == 2
        assert foo('a') == 1
        assert foo.pop('a') == 1
        assert foo('b') == 2
        assert foo('a') == 3

    def test_method(self):
        class Foo(object):
            def __init__(self):
                self.counter = 0

            @MemoizeDecorator()
            def foo(self, a):
                self.counter += 1
                return self.counter

        descriptor = Foo.__dict__['foo']

        assert isinstance(descriptor, MemoizeDescriptor)
        assert not descriptor.as_property



from __future__ import absolute_import, print_function, unicode_literals

import mock
import pytest
from django.core.exceptions import MiddlewareNotUsed
from django.http.response import HttpResponse

from django_auxilium.middleware import html
from django_auxilium.middleware.html import MinifyHTMLMiddleware


class TestMinifyHTMLMiddleware(object):
    def test_init(self, settings):
        settings.DEBUG = False

        assert isinstance(MinifyHTMLMiddleware(), MinifyHTMLMiddleware)

        settings.DEBUG = True
        with pytest.raises(MiddlewareNotUsed):
            MinifyHTMLMiddleware()

    @mock.patch.object(html, 'simple_minify')
    def test_process_response_html(self, mock_simple_minify, settings):
        settings.DEBUG = False
        response = HttpResponse('foo  ')
        response['Content-Type'] = 'text/html'
        mock_simple_minify.return_value = 'hello there'

        middleware = MinifyHTMLMiddleware()

        assert middleware.process_response(None, response) is response
        assert response.content == b'hello there'
        mock_simple_minify.assert_called_once_with(b'foo')

    @mock.patch.object(html, 'simple_minify')
    def test_process_response_error(self, mock_simple_minify, settings):
        settings.DEBUG = False
        response = HttpResponse('foo  ')
        response['Content-Type'] = 'text/html'
        mock_simple_minify.side_effect = ValueError

        middleware = MinifyHTMLMiddleware()

        assert middleware.process_response(None, response) is response
        assert response.content == b'foo  '
        mock_simple_minify.assert_called_once_with(b'foo')



from __future__ import absolute_import, print_function, unicode_literals

import magic
import mock

from django_auxilium.utils import content_type
from django_auxilium.utils.content_type import (
    get_content_type,
    get_content_type_from_binary,
    get_content_type_from_file,
)


@mock.patch.object(magic, 'from_buffer')
def test_get_content_type_from_binary(mock_from_buffer):
    mock_from_buffer.return_value = b'plain/text'

    actual = get_content_type_from_binary(b'foo')

    assert actual == 'plain/text'
    mock_from_buffer.assert_called_once_with(b'foo', mime=True)


@mock.patch.object(content_type, 'get_content_type_from_binary')
def test_get_content_type_from_file(mock_get_content_type_from_binary):
    f = mock.Mock()

    actual = get_content_type_from_file(f)

    assert actual == mock_get_content_type_from_binary.return_value
    assert f.mock_calls == [
        mock.call.seek(0),
        mock.call.read(1024),
        mock.call.seek(0),
    ]
    mock_get_content_type_from_binary.assert_called_once_with(f.read.return_value)


@mock.patch.object(content_type, 'get_content_type_from_file')
@mock.patch.object(content_type, 'get_content_type_from_binary')
def test_get_content_type_binary(mock_get_content_type_from_binary, mock_get_content_type_from_file):
    actual = get_content_type(b'foo')

    assert actual == mock_get_content_type_from_binary.return_value
    mock_get_content_type_from_binary.assert_called_once_with(b'foo')
    assert not mock_get_content_type_from_file.called


@mock.patch.object(content_type, 'get_content_type_from_file')
@mock.patch.object(content_type, 'get_content_type_from_binary')
def test_get_content_type_file(mock_get_content_type_from_binary, mock_get_content_type_from_file):
    actual = get_content_type(mock.sentinel.file)

    assert actual == mock_get_content_type_from_file.return_value
    mock_get_content_type_from_file.assert_called_once_with(mock.sentinel.file)
    assert not mock_get_content_type_from_binary.called



from __future__ import print_function, unicode_literals
import re

from django import forms
from django.utils.translation import ugettext_lazy as _

from django_auxilium.utils.range import AlphabeticNumbers, Range


class RangeSelectorField(forms.CharField):
    """
    Range field which supports Excel-type rangeconfig selectors.

    The rangeconfig selection is made using the following format::

        <Column><Row>:<Column><Row>

    Examples
    --------

    ::

        A:A
        A1:A50
        1:50

    Parameters
    ----------
    max_rows : int, optional
        The maximum number of rows the range can have
    max_columns : int, optional
        The maximum number of columns the range can have
    max_either : int, optional
        The maximum number of rows and columns the range can have.
        For example if the value is 1 then either a single column or
        a single row can be selected.
    required_rows : bool, optional
        Whether rows must be supplied in the rangeconfig
    required_columns : bool, optional
        Whether columns must be supplied in the rangeconfig
    """

    default_error_messages = {
        "invalid": _("Invalid range value."),
        "values": _("Top-left coordinate must be first."),
        "max_rows": _("Too many rows selected. Maximum is {0}."),
        "max_columns": _("Too many columns selected. Maximum is {0}."),
        "max_either": _("Too many rows or columns selected. Maximum is {0}."),
        "max_total": _("Too many total cells selected. Maximum is {0}."),
        "required_rows": _("Row selection is required."),
        "required_columns": _("Column selection is required."),
    }

    def __init__(self, *args, **kwargs):
        self.max_rows = kwargs.pop('max_rows', None)
        self.max_columns = kwargs.pop('max_columns', None)
        self.max_either = kwargs.pop('max_either', None)
        self.required_rows = kwargs.pop('required_rows', False)
        self.required_columns = kwargs.pop('required_columns', False)

        super(RangeSelectorField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        """
        Convert range string value to ``Range`` rangeconfig value.

        Returns
        -------
        Range
        """
        value = super(RangeSelectorField, self).to_python(value)
        if value in self.empty_values:
            return

        # extract all components
        r = re.compile(
            r'^(?P<A>(?P<A1>[A-Z]+):(?P<A2>[A-Z]+))$'
            r'|^(?P<N>(?P<N1>\d+):(?P<N2>\d+))$'
            r'|^(?P<AN>(?P<AN_A1>[A-Z]+)(?P<AN_N1>\d+):(?P<AN_A2>[A-Z]+)(?P<AN_N2>\d+))$'
        )

        # make sure overall pattern is valid
        if not r.findall(value):
            raise forms.ValidationError(self.error_messages['invalid'])

        datarange = None

        # extract rangeconfig values
        m = r.match(value)
        for group in ['A', 'N', 'AN']:
            if not m.group(group):
                continue
            if group == 'A':
                datarange = Range(
                    AlphabeticNumbers.int_from_str(m.group('A1')),
                    None,
                    AlphabeticNumbers.int_from_str(m.group('A2')),
                    None
                )
            elif group == 'N':
                datarange = Range(
                    None,
                    int(m.group('N1')),
                    None,
                    int(m.group('N2'))
                )
            else:
                datarange = Range(
                    AlphabeticNumbers.int_from_str(m.group('AN_A1')),
                    int(m.group('AN_N1')),
                    AlphabeticNumbers.int_from_str(m.group('AN_A2')),
                    int(m.group('AN_N2'))
                )

        return datarange

    def validate(self, value):
        """
        Validate the rangeconfig value.

        Following is validated:

        * rows or columns are supplied depending on ``required_row``
          and ``required_columns`` parameters
        * rangeconfig is given as top-left to bottom-right
        * columns range is within ``max_columns``
        * rows range is within ``max_rows``
        * both columns and rows range is within ``max_either``
        """
        if value in self.empty_values:
            return

        if self.required_columns and not value.first_column:
            raise forms.ValidationError(
                self.error_messages['required_columns']
            )
        if self.required_rows and not value.first_row:
            raise forms.ValidationError(
                self.error_messages['required_rows']
            )

        # make sure the first rangeconfig coordinate is top-left
        for i, j in zip(value[:2], value[2:]):
            if i and j and i > j:
                raise forms.ValidationError(self.error_messages['values'])

        if self.max_columns and value.columns > self.max_columns:
            raise forms.ValidationError(
                self.error_messages['max_columns'].format(self.max_columns)
            )

        if self.max_rows and value.rows > self.max_rows:
            raise forms.ValidationError(
                self.error_messages['max_rows'].format(self.max_rows)
            )

        if self.max_either:
            if all([value.columns > self.max_either,
                    value.rows > self.max_either]):
                raise forms.ValidationError(
                    self.error_messages['max_either'].format(self.max_either)
                )



from __future__ import print_function, unicode_literals
import random
import re

import pytest
import six
from django import forms

from django_auxilium.forms.multiple_values import MultipleValuesCharField


class TestMultipleValuesCharField(object):
    def test_empty(self):
        field = MultipleValuesCharField(required=False)
        data = ['', None]
        for d in data:
            assert field.clean(d) == []

    def test_delimiter(self):
        delimiters = [',', ';', ':']
        data = ['foo', 'bar', 'foo2', 'bar2']
        for i in delimiters:
            field = MultipleValuesCharField(delimiter=i)
            value = i.join(data)
            assert field.clean(value) == data

        delimiter = re.compile('\W+')
        data = 'hello world, how  are you?'
        expected = ['hello', 'world', 'how', 'are', 'you']
        field = MultipleValuesCharField(delimiter=delimiter)
        assert field.clean(data) == expected

    def test_mapping(self):
        mapping = {
            '1': 1
        }

        def mapping_callable(value):
            if value == '1':
                return 1
            else:
                return value

        data = ['foo', '1', 'bar']
        expected = ['foo', 1, 'bar']
        value = ','.join(data)

        field = MultipleValuesCharField(mapping=mapping)
        assert field.clean(value) == expected

        field = MultipleValuesCharField(mapping=mapping_callable)
        assert field.clean(value) == expected

        field = MultipleValuesCharField(mapping=int)
        expected = random.sample(six.moves.range(100000), 5)
        value = ','.join([six.text_type(i) for i in expected])
        assert field.clean(value) == expected

        with pytest.raises(forms.ValidationError):
            field.clean('foo')

    def test_max_values(self):
        expected = ['foo', 'bar', 'foo2', 'bar2']
        data = ','.join(expected)

        field = MultipleValuesCharField(max_values=2)
        with pytest.raises(forms.ValidationError):
            field.clean(data)
        with pytest.raises(forms.ValidationError):
            field.clean('')

        field = MultipleValuesCharField(max_length=10)
        assert field.clean(data) == expected

        field = MultipleValuesCharField(max_length=10, required=False)
        assert field.clean('') == []

    def test_min_values(self):
        expected = ['foo', 'bar', 'foo2', 'bar2']
        data = ','.join(expected)

        field = MultipleValuesCharField(min_values=10)
        with pytest.raises(forms.ValidationError):
            field.clean(data)
        with pytest.raises(forms.ValidationError):
            field.clean('')

        field = MultipleValuesCharField(min_length=2)
        assert field.clean(data) == expected

        field = MultipleValuesCharField(min_length=2, required=False)
        assert field.clean('') == []

    def test_strip(self):
        data = ['foo', 'bar', 'foo2   ', '   bar2']
        true_expected = [i.strip() for i in data]
        false_expected = data
        value = ','.join(data)

        field = MultipleValuesCharField(strip=True)
        assert field.clean(value) == true_expected

        field = MultipleValuesCharField(strip=False)
        assert field.clean(value) == false_expected

    def test_disregard_empty(self):
        value = 'foo,,    ,bar,'
        true_expected = ['foo', '    ', 'bar']
        false_expected = ['foo', '', '    ', 'bar', '']

        field = MultipleValuesCharField(disregard_empty=True, strip=False)
        assert field.clean(value) == true_expected

        field = MultipleValuesCharField(disregard_empty=False, strip=False)
        assert field.clean(value) == false_expected

    def test_invalid_values(self):
        data = ['foo', 'bar', 'foo2']
        value = ','.join(data)
        invalid = ['bar', 'foo2']

        field = MultipleValuesCharField(invalid_values=invalid)
        with pytest.raises(forms.ValidationError):
            field.clean(value)

    def test_prepare_value(self):
        field = MultipleValuesCharField()
        assert field.prepare_value('foo') == 'foo'
        assert field.prepare_value(['foo', 'bar']) == 'foo,bar'



from __future__ import print_function, unicode_literals

from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed

from django_auxilium.utils.html import simple_minify


class MinifyHTMLMiddleware(object):
    """
    Middleware for minifying HTML.

    Unlike some other minifyers, this minifyer is pretty
    simple and tried to do the bare minimum in an attempt
    of being super-lightweight.

    This middleware only works when ``settings.DEBUG``
    is disabled.

    See Also
    --------
    django_auxilium.utils.html.simple_minify
        What is used to actually minify HTML
    """

    def __init__(self):
        if settings.DEBUG:
            raise MiddlewareNotUsed

    def process_response(self, request, response):
        """
        Process the response by minifying HTML.

        HTML is minified only when the response content type
        is HTML (e.g. contains ``'text/html'``).
        """
        header = 'Content-Type'
        if response.has_header(header) and 'text/html' in response[header]:
            try:
                response.content = simple_minify(response.content.strip())
            except Exception:
                pass
        return response



from __future__ import unicode_literals

from .signals import failed_mail, sending_mail, sent_mail


class BaseCallbackEmailBackendMixin(object):
    """
    Base class for implementing callback email backends.

    "Callback" is used in a sense that appropriate signals
    are sent while sending message.

    :Sending: :py:data:`.signals.sending_mail` signal is sent when
              message is in progress of being sent
    :Sent:    :py:data:`.signals.sent_mail` signal is sent when
              message was successfully sent
    :Failed:  :py:data:`.signals.failed_mail` signal is sent when
              message sending failed
    """

    def sending(self, message):
        """
        Hook which sends signal when message is being sent
        """
        sending_mail.send(sender=self.__class__, message=message)

    def sent(self, message):
        """
        Hook which sends signal when message was successfully sent
        """
        sent_mail.send(sender=self.__class__, message=message)

    def failed(self, message, reason=None):
        """
        Hook which sends signal when sending message failed
        """
        failed_mail.send(
            sender=self.__class__,
            message=message,
            reason=reason
        )



from __future__ import unicode_literals

import six


def get_content_type_from_binary(data):
    """
    Get content_type of a binary string by using ``magic`` library

    Parameters
    ----------
    data : binary
        Binary string data of a file from which
        content_type will be determined

    Returns
    -------
    str
        Content-type of a file as string
    """
    # has to be inline import since this lib does not
    # required python-magic by default.
    import magic

    return magic.from_buffer(data, mime=True).decode('utf-8')


def get_content_type_from_file(f):
    """
    Get content_type of a file-like object by using ``magic`` library

    .. note::
        This function reads first 1Kb of the file in order
        to determine the content_type however at the end
        seeks back to ``0``

    Parameters
    ----------
    f : file
        File-like object which should implement
        ``.seek()`` and ``.read()`` methods
        from which content_type will be determined

    Returns
    -------
    str
        Content-type of a file as string
    """
    f.seek(0)
    try:
        return get_content_type_from_binary(f.read(1024))
    finally:
        f.seek(0)


def get_content_type(f):
    """
    Get content_type of a file by using ``magic`` library

    Parameters
    ----------
    f : str or file
        Either a binary string of a file content or a
        file-like object from which file data will be read

    Returns
    -------
    str
        Content-type of a file as string


    See Also
    --------
    get_content_type_from_binary
    get_content_type_from_file
    """
    if isinstance(f, six.binary_type):
        return get_content_type_from_binary(f)
    else:
        return get_content_type_from_file(f)



from __future__ import print_function, unicode_literals
import re

import six
from django.utils.html import strip_spaces_between_tags
from six.moves.html_entities import name2codepoint
from six.moves.html_parser import HTMLParser

from .string import text_onion


EXCLUDE_TAGS = ('textarea', 'pre', 'code', 'script',)
RE_WHITESPACE = re.compile(r'\s{2,}|\n')
RE_EXCLUDE_TAGS = re.compile(
    """(               # group for results to be included in re.split
          <(?:{0})     # match beginning of one of exclude tags
                       #   e.g. <pre or <textarea
          .*?          # non-greedy match anything inside the tag
                       # until the next group is matched
          </(?:{0})>   # match the closing tag   e.g. </pre>
    )""".format('|'.join(EXCLUDE_TAGS)),
    re.DOTALL | re.VERBOSE)


@text_onion
def simple_minify(html):
    """
    Minify HTML with very simple algorithm.

    This function tries to minify HTML by stripping all spaces between all html tags
    (e.g. ``</div>    <div>`` -> ``</div><div>``). This step is accomplished by using
    Django's ``strip_spaces_between_tags`` method. In addition to that, this function
    replaces all whitespace (more then two consecutive whitespace characters or new line)
    with a space character except inside excluded tags such as ``pre`` or ``textarea``.

    **Though process**:

    To minify everything except content of excluded tags in one step requires very
    complex regular expression. The disadvantage is the regular expression will involve
    look-behinds and look-aheads. Those operations make regex much more resource-hungry
    which eats precious server resources. In addition, complex regex are hard to understand
    and can be hard to maintain. That is why this function splits the task into multiple
    sections.

    #. Regex expression which matches all exclude tags within the html is used
       to split the HTML split into components. Since the regex expression is
       wrapped inside a group, the content of the exclude tags is also included
       inside the resulting split list.
       Due to that it is guaranteed that every odd element (if there are any)
       will be the excluded tags.
    #. All the split components are looped and processed in order to construct
       final minified HTML.
    #. All odd indexed elements are not processed and are simply
       appended to final HTML since as explained above, they are guaranteed
       to be content of excluded tags hence do not require minification.
    #. All even indexed elements are minified by stripping whitespace between
       tags by using Django's ``strip_spaces_between_tags`` and redundant
       whitespace is stripped in general via simple regex.

    You can notice that the process does not involve parsing HTML since that
    usually adds some overhead (e.g. using beautiful soup). By using 2 regex
    passes this achieves very similar result which performs much better.
    """
    components = RE_EXCLUDE_TAGS.split(html)
    html = ''
    for i, component in enumerate(components):
        if i % 2 == 0:
            component = strip_spaces_between_tags(component.strip())
            component = RE_WHITESPACE.sub(' ', component)
            html += component
        else:
            html += component
    return html


class TextExtractorHTMLParser(HTMLParser):
    """
    Custom HTML parser which extracts only text while parsing HTML

    Once the parser parses the HTML, :py:meth:`.get_text`
    can be called which will return extracted text
    """

    def __init__(self):
        HTMLParser.__init__(self)
        self.result = []

    def handle_data(self, d):
        """
        Handler for data/text in HTML

        This simply adds the data to the results list this class
        maintains of the extracted html
        """
        self.result.append(d)

    def handle_charref(self, number):
        """
        Handler for processing character references.

        This method handles both decimal (e.g. ``'&gt;' == '&#62;'``) and
        hexadecimal (e.g. ``'&gt;' == '&#x3E;'``) references.
        It does that by simply converting the reference number
        to an integer with appropriate base and then converts
        that number to a character.
        """
        codepoint = int(number[1:], 16) if number[0] in ('x', 'X') else int(number)
        text = six.unichr(codepoint)
        self.result.append(text)
        return text

    def handle_entityref(self, name):
        """
        Handler for processing character references.

        This method handles processing HTML entities (e.g. ``'&gt;'``).
        It first maps the entity name to a codepoint which is a
        unicode character number and then converts that number
        to a unicode character.
        """
        text = six.unichr(name2codepoint[name])
        self.result.append(text)
        return text

    def get_text(self):
        """
        Get extracted text after HTML is parsed.

        Returns
        -------
        str
            Extracted text from the HTML document
        """
        return ''.join(self.result)


def html_to_text(html):
    """
    Function to convert HTML text to plain text by stripping all HTML.

    Implementation is based from
    `<http://stackoverflow.com/questions/753052/strip-html-from-strings-in-python>`_.
    """
    s = TextExtractorHTMLParser()
    s.feed(html)
    return s.get_text()



from __future__ import print_function, unicode_literals


try:
    from formtools.wizard.views import WizardView as _WizardView  # >=django1.8
except ImportError:
    from django.contrib.formtools.wizard.views import WizardView as _WizardView  # <django1.8


class WizardView(_WizardView):
    """
    Modified Django's ``WizardView`` which allows to specify ``form_list`` as
    class attribute.

    When using Django's ``WizardView``, the ``form_list`` must be specified
    when calling ``as_view()``::

        WizardView.as_view(form_list)

    This class allows to provide the ``form_list`` as class attribute::

        Wizard(WizardView):
            form_list = []

        Wizard.as_view()

    .. warning::
        This class requires `django-formtools <http://django-formtools.readthedocs.org/en/latest/>`_
        to be independently installed since this library does not install
        it as a dependency.
    """
    form_list = []

    @classmethod
    def get_initkwargs(cls, *args, **kwargs):
        """
        Custom ``get_initkwargs`` which adds ``form_list``
        to the super call when class defines any forms.
        """
        if cls.form_list:
            args = [cls.form_list] + list(args)
        return super(WizardView, cls).get_initkwargs(*args, **kwargs)
